# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import ir


def _is_scan(node: ir.Node):
    return (
        isinstance(node, ir.FunCall)
        and isinstance(node.fun, ir.FunCall)
        and node.fun.fun == ir.SymRef(id="scan")
    )


class ScanEtaReduction(PreserveLocationVisitor, NodeTranslator):
    """Applies eta-reduction-like transformation involving scans.

    Simplifies `λ(x, y) → scan(λ(state, param_y, param_x) → ..., ...)(y, x)` to `scan(λ(state, param_x, param_y) → ..., ...)`.
    Note, unlike the eta reduction, this pass works even if parameters of the lambda and arguments in the call
    of the scanpass are not in the same order: parameters of the scanpass are re-ordered. This limits the pass
    to lambdas, i.e. doesn't apply to builtins and `FunctionDefinition`.
    """

    def visit_Lambda(self, node: ir.Lambda) -> ir.Node:
        if _is_scan(node.expr):
            assert isinstance(node.expr, ir.FunCall)
            if len(node.params) == len(node.expr.args) and all(
                isinstance(a, ir.SymRef) and p.id == a.id
                for p, a in zip(
                    sorted(node.params, key=lambda x: str(x)),
                    sorted(node.expr.args, key=lambda x: str(x)),
                )
            ):
                # node.expr.fun is the unapplied scan
                assert isinstance(node.expr.fun, ir.FunCall)
                original_scanpass = node.expr.fun.args[0]
                assert isinstance(original_scanpass, ir.Lambda)
                new_scanpass_params_idx = []
                args_str = [str(a) for a in node.expr.args]
                for p in (str(p) for p in node.params):
                    new_scanpass_params_idx.append(args_str.index(p))
                new_scanpass_params = [original_scanpass.params[0]] + [
                    original_scanpass.params[i + 1] for i in new_scanpass_params_idx
                ]
                new_scanpass = ir.Lambda(params=new_scanpass_params, expr=original_scanpass.expr)
                return ir.FunCall(
                    fun=ir.SymRef(id="scan"), args=[new_scanpass, *node.expr.fun.args[1:]]
                )

        return self.generic_visit(node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py import eve
from gt4py.next.iterator import builtins, ir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm
from gt4py.next.type_system import type_specifications as ts


class PruneCasts(eve.NodeTranslator):
    """
    Removes cast expressions where the argument is already in the target type.

    This transformation requires the IR to be fully type-annotated,
    therefore it should be applied after type-inference.
    """

    PRESERVED_ANNEX_ATTRS = ("domain",)

    def visit_FunCall(self, node: ir.FunCall) -> ir.Node:
        node = self.generic_visit(node)

        if cpm.is_call_to(node, "cast_"):
            value, type_constructor = node.args

            assert (
                value.type
                and isinstance(type_constructor, ir.SymRef)
                and (type_constructor.id in builtins.TYPE_BUILTINS)
            )
            dtype = ts.ScalarType(kind=getattr(ts.ScalarKind, type_constructor.id.upper()))

            if value.type == dtype:
                return value

        elif cpm.is_identity_as_fieldop(node):
            # pruning of cast expressions may leave some trivial `as_fieldop` expressions
            # with form '(⇑(λ(__arg) → ·__arg))(a)'
            return node.args[0]

        return node

    @classmethod
    def apply(cls, node: ir.Node) -> ir.Node:
        return cls().visit(node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Contains type definitions and data utilities used during lowering to SDFG."""

from __future__ import annotations

import dataclasses
from typing import Final, TypeAlias

import dace
from dace import subsets as dace_subsets

from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.ffront import fbuiltins as gtx_fbuiltins
from gt4py.next.program_processors.runners.dace import (
    gtir_dataflow,
    gtir_domain,
    gtir_to_sdfg,
    gtir_to_sdfg_utils,
    utils as gtx_dace_utils,
)
from gt4py.next.type_system import type_specifications as ts


@dataclasses.dataclass(frozen=True)
class FieldopData:
    """
    Abstraction to represent data (scalars, arrays) during the lowering to SDFG.

    Attribute 'local_offset' must always be set for `FieldType` data with a local
    dimension generated from neighbors access in unstructured domain, and indicates
    the name of the offset provider used to generate the list of neighbor values.

    Args:
        dc_node: DaCe access node to the data storage.
        gt_type: GT4Py type definition, which includes the field domain information.
        origin: Tuple of start indices, in each dimension, for `FieldType` data.
            Pass an empty tuple for `ScalarType` data or zero-dimensional fields.
    """

    dc_node: dace.nodes.AccessNode
    gt_type: ts.FieldType | ts.ScalarType
    origin: tuple[dace.symbolic.SymbolicType, ...]

    def __post_init__(self) -> None:
        """Implements a sanity check on the constructed data type."""
        assert (
            len(self.origin) == 0
            if isinstance(self.gt_type, ts.ScalarType)
            else len(self.origin) == len(self.gt_type.dims)
        )

    def map_to_parent_sdfg(
        self,
        sdfg_builder: gtir_to_sdfg.SDFGBuilder,
        inner_sdfg: dace.SDFG,
        outer_sdfg: dace.SDFG,
        outer_sdfg_state: dace.SDFGState,
        symbol_mapping: dict[str, dace.symbolic.SymbolicType],
    ) -> FieldopData:
        """
        Make the data descriptor which 'self' refers to, and which is located inside
        a NestedSDFG, available in its parent SDFG.

        Thus, it turns 'self' into a non-transient array and creates a new data
        descriptor inside the parent SDFG, with same shape and strides.
        """
        inner_desc = self.dc_node.desc(inner_sdfg)
        assert inner_desc.transient
        inner_desc.transient = False

        if isinstance(self.gt_type, ts.ScalarType):
            outer, outer_desc = sdfg_builder.add_temp_scalar(outer_sdfg, inner_desc.dtype)
            outer_origin = []
        else:
            outer, outer_desc = sdfg_builder.add_temp_array_like(outer_sdfg, inner_desc)
            # We cannot use a copy of the inner data descriptor directly, we have to apply the symbol mapping.
            dace.symbolic.safe_replace(
                symbol_mapping,
                lambda m: dace.sdfg.replace_properties_dict(outer_desc, m),
            )
            # Same applies to the symbols used as field origin (the domain range start)
            outer_origin = [
                gtx_dace_utils.safe_replace_symbolic(val, symbol_mapping) for val in self.origin
            ]

        outer_node = outer_sdfg_state.add_access(outer)
        return FieldopData(outer_node, self.gt_type, tuple(outer_origin))

    def get_local_view(
        self, domain: gtir_domain.FieldopDomain, sdfg: dace.SDFG
    ) -> gtir_dataflow.IteratorExpr | gtir_dataflow.MemletExpr:
        """Helper method to access a field in local view, given the compute domain of a field operator."""
        if isinstance(self.gt_type, ts.ScalarType):
            assert isinstance(self.dc_node.desc(sdfg), dace.data.Scalar)
            return gtir_dataflow.MemletExpr(
                dc_node=self.dc_node,
                gt_dtype=self.gt_type,
                subset=dace_subsets.Range.from_string("0"),
            )

        if isinstance(self.gt_type, ts.FieldType):
            it_indices: dict[gtx_common.Dimension, gtir_dataflow.DataExpr]
            if isinstance(self.dc_node.desc(sdfg), dace.data.Scalar):
                assert len(self.gt_type.dims) == 0  # zero-dimensional field
                it_indices = {}
            else:
                # The invariant below is ensured by calling `make_field()` to construct `FieldopData`.
                # The `make_field` constructor converts any local dimension, if present, to `ListType`
                # element type, while leaving the field domain with all global dimensions.
                assert all(dim != gtx_common.DimensionKind.LOCAL for dim in self.gt_type.dims)
                domain_dims = [domain_range.dim for domain_range in domain]
                domain_indices = gtir_domain.get_domain_indices(domain_dims, origin=None)
                it_indices = {
                    dim: gtir_dataflow.SymbolExpr(index, INDEX_DTYPE)
                    for dim, index in zip(domain_dims, domain_indices)
                }
            field_origin = [
                (dim, dace.symbolic.SymExpr(0) if self.origin is None else self.origin[i])
                for i, dim in enumerate(self.gt_type.dims)
            ]
            return gtir_dataflow.IteratorExpr(
                self.dc_node, self.gt_type.dtype, field_origin, it_indices
            )

        raise NotImplementedError(f"Node type {type(self.gt_type)} not supported.")

    def get_symbol_mapping(
        self, dataname: str, sdfg: dace.SDFG
    ) -> dict[str, dace.symbolic.SymExpr]:
        """
        Helper method to create the symbol mapping for array storage in a nested SDFG.

        Args:
            dataname: Name of the data container insiode the nested SDFG.
            sdfg: The parent SDFG where the `FieldopData` object lives.

        Returns:
            Mapping from symbols in nested SDFG to the corresponding symbolic values
            in the parent SDFG. This includes the range start and stop symbols (used
            to calculate the array shape as range 'stop - start') and the strides.
        """
        if isinstance(self.gt_type, ts.ScalarType):
            return {}
        ndims = len(self.gt_type.dims)
        outer_desc = self.dc_node.desc(sdfg)
        assert isinstance(outer_desc, dace.data.Array)
        # origin and size of the local dimension, in case of a field with `ListType` data,
        # are assumed to be compiled-time values (not symbolic), therefore the start and
        # stop range symbols of the inner field only extend over the global dimensions
        return (
            {gtx_dace_utils.range_start_symbol(dataname, i): (self.origin[i]) for i in range(ndims)}
            | {
                gtx_dace_utils.range_stop_symbol(dataname, i): (
                    self.origin[i] + outer_desc.shape[i]
                )
                for i in range(ndims)
            }
            | {
                gtx_dace_utils.field_stride_symbol_name(dataname, i): stride
                for i, stride in enumerate(outer_desc.strides)
            }
        )


FieldopResult: TypeAlias = FieldopData | tuple[FieldopData | tuple, ...]
"""Result of a field operator, can be either a field or a tuple fields."""


@dataclasses.dataclass(frozen=True)
class SymbolicData:
    gt_type: ts.ScalarType
    value: dace.symbolic.SymbolicType


INDEX_DTYPE: Final[dace.typeclass] = dace.dtype_to_typeclass(gtx_fbuiltins.IndexType)
"""Data type used for field indexing."""


def get_tuple_type(data: tuple[FieldopResult, ...]) -> ts.TupleType:
    """
    Compute the `ts.TupleType` corresponding to the tuple structure of `FieldopResult`.
    """
    return ts.TupleType(
        types=[get_tuple_type(d) if isinstance(d, tuple) else d.gt_type for d in data]
    )


def flatten_tuples(name: str, arg: FieldopResult) -> list[tuple[str, FieldopData]]:
    """
    Visit a `FieldopResult`, potentially containing nested tuples, and construct a list
    of pairs `(str, FieldopData)` containing the symbol name of each tuple field and
    the corresponding `FieldopData`.
    """
    if isinstance(arg, tuple):
        tuple_type = get_tuple_type(arg)
        tuple_symbols = gtir_to_sdfg_utils.flatten_tuple_fields(name, tuple_type)
        tuple_data_fields = gtx_utils.flatten_nested_tuple(arg)
        return [
            (str(tsym.id), tfield)
            for tsym, tfield in zip(tuple_symbols, tuple_data_fields, strict=True)
        ]
    else:
        return [(name, arg)]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import typing
from typing import Any, Generic, Iterable, Iterator, Optional

from typing_extensions import Self

from gt4py._core import definitions as core_defs
from gt4py.next import common
from gt4py.next.otf import toolchain, workflow
from gt4py.next.type_system import type_info, type_specifications as ts, type_translation


DATA_T = typing.TypeVar("DATA_T")


@dataclasses.dataclass(frozen=True)
class StaticArg(Generic[core_defs.ScalarT]):
    value: core_defs.ScalarT | tuple[core_defs.ScalarT | tuple, ...]
    type_: ts.TypeSpec


@dataclasses.dataclass(frozen=True)
class JITArgs:
    """Concrete (runtime) arguments to a GTX program in a format that can be passed into the toolchain."""

    args: tuple[Any, ...]
    kwargs: dict[str, Any]

    @classmethod
    def from_signature(cls, *args: Any, **kwargs: Any) -> Self:
        return cls(args=args, kwargs=kwargs)


@dataclasses.dataclass(frozen=True)
class CompileTimeArgs:
    """Compile-time standins for arguments to a GTX program to be used in ahead-of-time compilation."""

    args: tuple[ts.TypeSpec | StaticArg, ...]
    kwargs: dict[str, ts.TypeSpec | StaticArg]
    offset_provider: common.OffsetProvider  # TODO(havogt): replace with common.OffsetProviderType once the temporary pass doesn't require the runtime information
    column_axis: Optional[common.Dimension]

    @property
    def offset_provider_type(self) -> common.OffsetProviderType:
        return common.offset_provider_to_type(self.offset_provider)

    @classmethod
    def from_concrete_no_size(cls, *args: Any, **kwargs: Any) -> Self:
        """Convert concrete GTX program arguments into their compile-time counterparts."""
        compile_args = tuple(type_translation.from_value(arg) for arg in args)
        kwargs_copy = kwargs.copy()
        offset_provider = kwargs_copy.pop("offset_provider", {})
        return cls(
            args=compile_args,
            offset_provider=offset_provider,
            column_axis=kwargs_copy.pop("column_axis", None),
            kwargs={
                k: type_translation.from_value(v) for k, v in kwargs_copy.items() if v is not None
            },
        )

    @classmethod
    def from_concrete(cls, *args: Any, **kwargs: Any) -> Self:
        """Convert concrete GTX program arguments to compile-time, adding (compile-time) dimension size arguments."""
        no_size = cls.from_concrete_no_size(*args, **kwargs)
        return cls(
            args=(*no_size.args, *iter_size_compile_args(no_size.args)),
            offset_provider=no_size.offset_provider,
            column_axis=no_size.column_axis,
            kwargs=no_size.kwargs,
        )

    @classmethod
    def empty(cls) -> Self:
        return cls(tuple(), {}, {}, None)


def jit_to_aot_args(
    inp: JITArgs,
) -> CompileTimeArgs:
    return CompileTimeArgs.from_concrete_no_size(*inp.args, **inp.kwargs)


def adapted_jit_to_aot_args_factory() -> workflow.Workflow[
    toolchain.CompilableProgram[DATA_T, JITArgs],
    toolchain.CompilableProgram[DATA_T, CompileTimeArgs],
]:
    """Wrap `jit_to_aot` into a workflow adapter to fit into backend transform workflows."""
    return toolchain.ArgsOnlyAdapter(jit_to_aot_args)


def find_first_field(tuple_arg: tuple[Any, ...]) -> Optional[common.Field]:
    for element in tuple_arg:
        match element:
            case tuple():
                found = find_first_field(element)
                if found:
                    return found
            case common.Field():
                return element
            case _:
                pass
    return None


def iter_size_args(args: tuple[Any, ...]) -> Iterator[tuple[int, int]]:
    """
    Yield the size of each field argument in each dimension.

    This can be used to generate domain size arguments for FieldView Programs that use an implicit domain.
    """
    for arg in args:
        match arg:
            case tuple():
                # we only need the first field, because all fields in a tuple must have the same dims and sizes
                first_field = find_first_field(arg)
                if first_field:
                    yield from iter_size_args((first_field,))
            case common.Field():
                for range_ in arg.domain.ranges:
                    assert isinstance(range_, common.UnitRange)
                    yield (range_.start, range_.stop)
            case _:
                pass


def iter_size_compile_args(
    args: Iterable[ts.TypeSpec | StaticArg],
) -> Iterator[ts.TypeSpec]:
    """
    Yield a compile-time size argument for every compile-time field argument in each dimension.

    This can be used inside transformation workflows to generate compile-time domain size arguments for FieldView Programs that use an implicit domain.
    """
    for arg in args:
        type_ = arg.type_ if isinstance(arg, StaticArg) else arg
        field_constituents: list[ts.FieldType] = typing.cast(
            list[ts.FieldType],
            type_info.primitive_constituents(type_).if_isinstance(ts.FieldType).to_list(),
        )
        if field_constituents:
            # we only need the first field, because all fields in a tuple must have the same dims and sizes
            index_type = ts.ScalarType(kind=ts.ScalarKind.INT32)
            yield from [
                ts.TupleType(types=[index_type, index_type]) for _ in field_constituents[0].dims
            ]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Utility functions for formatting :class:`DSLError` and its subclasses."""

from __future__ import annotations

import linecache
import textwrap
import traceback
import types
from typing import TYPE_CHECKING, Optional

from gt4py.eve import SourceLocation


if TYPE_CHECKING:
    from . import exceptions


def get_source_from_location(location: SourceLocation) -> str:
    if not location.filename:
        raise FileNotFoundError()
    source_lines = linecache.getlines(location.filename)
    if not source_lines:
        raise FileNotFoundError()
    start_line = location.line
    end_line = location.end_line + 1 if location.end_line is not None else start_line + 1
    relevant_lines = source_lines[(start_line - 1) : (end_line - 1)]
    return "\n".join(relevant_lines)


def format_location(loc: SourceLocation, show_caret: bool = False) -> str:
    """
    Format the source file location.

    Args:
        show_caret (bool): Indicate the position within the source line by placing carets underneath.
    """
    filename = loc.filename or "<unknown>"
    lineno = loc.line
    loc_str = f'File "{filename}", line {lineno}'

    if show_caret and loc.column is not None:
        offset = loc.column - 1
        width = loc.end_column - loc.column if loc.end_column is not None else 1
        caret_str = "".join([" "] * offset + ["^"] * width)
    else:
        caret_str = None

    try:
        snippet_str = get_source_from_location(loc)
        if caret_str:
            snippet_str = f"{snippet_str}{caret_str}"
        return f"{loc_str}\n{textwrap.indent(snippet_str, '  ')}"
    except (FileNotFoundError, IndexError):
        return loc_str


def _format_cause(cause: BaseException) -> list[str]:
    """Format the cause of an exception plus the bridging message to the current exception."""
    bridging_message = "The above exception was the direct cause of the following exception:"
    cause_strs = [*traceback.format_exception(cause), "\n", f"{bridging_message}\n\n"]
    return cause_strs


def _format_traceback(tb: types.TracebackType) -> list[str]:
    """Format the traceback of an exception."""
    intro_message = "Traceback (most recent call last):"
    traceback_strs = [f"{intro_message}\n", *traceback.format_tb(tb)]
    return traceback_strs


def format_compilation_error(
    type_: type[exceptions.DSLError],
    message: str,
    location: Optional[SourceLocation],
    tb: Optional[types.TracebackType] = None,
    cause: Optional[BaseException] = None,
) -> list[str]:
    bits: list[str] = []

    if cause is not None:
        bits = [*bits, *_format_cause(cause)]
    if tb is not None:
        bits = [*bits, *_format_traceback(tb)]
    if location is not None:
        loc_str = format_location(location, show_caret=True)
        loc_str_all = f"Source location:\n{textwrap.indent(loc_str, '  ')}\n"
        bits = [*bits, loc_str_all]
    msg_str = f"{type_.__module__}.{type_.__name__}: {message}"
    bits = [*bits, msg_str]
    return bits



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import enum
import functools
import operator
import re
from typing import Literal, Optional

from gt4py import eve
from gt4py.eve import utils as eve_utils
from gt4py.next import common
from gt4py.next.iterator import ir, ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    ir_makers as im,
    misc as ir_misc,
)
from gt4py.next.iterator.transforms import fixed_point_transformation, inline_lambdas, inline_lifts
from gt4py.next.iterator.type_system import (
    inference as itir_type_inference,
    type_specifications as it_ts,
)
from gt4py.next.type_system import type_info, type_specifications as ts


def _with_altered_iterator_element_type(
    type_: it_ts.IteratorType, new_el_type: ts.DataType
) -> it_ts.IteratorType:
    return it_ts.IteratorType(
        position_dims=type_.position_dims, defined_dims=type_.defined_dims, element_type=new_el_type
    )


def _with_altered_iterator_position_dims(
    type_: it_ts.IteratorType, new_position_dims: list[common.Dimension] | Literal["unknown"]
) -> it_ts.IteratorType:
    return it_ts.IteratorType(
        position_dims=new_position_dims,
        defined_dims=type_.defined_dims,
        element_type=type_.element_type,
    )


def _is_trivial_make_tuple_call(node: itir.Expr):
    """Return if node is a `make_tuple` call with all elements `SymRef`s, `Literal`s or tuples thereof."""
    if not cpm.is_call_to(node, "make_tuple"):
        return False
    if not all(_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args):
        return False
    return True


def _is_trivial_or_tuple_thereof_expr(node: itir.Node) -> bool:
    """
    Return `true` if the expr is a trivial expression (`SymRef` or `Literal`) or tuple thereof.

    Let forms with trivial body and args as well as `if` calls with trivial branches are also
    considered trivial.

    >>> _is_trivial_or_tuple_thereof_expr(im.make_tuple("a", "b"))
    True
    >>> _is_trivial_or_tuple_thereof_expr(im.tuple_get(1, "a"))
    True
    >>> _is_trivial_or_tuple_thereof_expr(
    ...     im.let("t", im.make_tuple("a", "b"))(im.tuple_get(1, "t"))
    ... )
    True
    """
    if isinstance(node, (itir.SymRef, itir.Literal)):
        return True
    if cpm.is_call_to(node, "make_tuple"):
        return all(_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args)
    if cpm.is_call_to(node, "tuple_get"):
        return _is_trivial_or_tuple_thereof_expr(node.args[1])
    # This will duplicate the condition and increase the size of the tree, but this is probably
    # acceptable.
    if cpm.is_call_to(node, "if_"):
        return all(_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args[1:])
    if cpm.is_let(node):
        return _is_trivial_or_tuple_thereof_expr(node.fun.expr) and all(
            _is_trivial_or_tuple_thereof_expr(arg) for arg in node.args
        )
    return False


def _flattened_as_fieldop_param_el_name(param: str, idx: int) -> str:
    prefix = "__ct_flat_el_"

    # keep the original param name, but skip prefix from previous flattenings
    if param.startswith(prefix):
        parent_idx, suffix = re.split(r"_(?!\d)", param[len(prefix) :], maxsplit=1)
        prefix = f"{prefix}{parent_idx}_"
    else:
        suffix = param

    return f"{prefix}{idx}_{suffix}"


# TODO(tehrengruber): Conceptually the structure of this pass makes sense: Visit depth first,
#  transform each node until no transformations apply anymore, whenever a node is to be transformed
#  go through all available transformation and apply them. However the final result here still
#  reads a little convoluted and is also different to how we write other transformations. We
#  should revisit the pattern here and try to find a more general mechanism.
@dataclasses.dataclass(frozen=True, kw_only=True)
class CollapseTuple(
    fixed_point_transformation.CombinedFixedPointTransform, eve.PreserveLocationVisitor
):
    """
    Simplifies `make_tuple`, `tuple_get` calls.

      - `make_tuple(tuple_get(0, t), tuple_get(1, t), ..., tuple_get(N-1,t))` -> `t`
      - `tuple_get(i, make_tuple(e_0, e_1, ..., e_i, ..., e_N))` -> `e_i`
    """

    # TODO(tehrengruber): This Flag mechanism is a little low level. What we actually want
    #   is something like a pass manager, where for each pattern we have a corresponding
    #   transformation, etc.
    class Transformation(enum.Flag):
        #: `make_tuple(tuple_get(0, t), tuple_get(1, t), ..., tuple_get(N-1,t))` -> `t`
        COLLAPSE_MAKE_TUPLE_TUPLE_GET = enum.auto()
        #: `tuple_get(i, make_tuple(e_0, e_1, ..., e_i, ..., e_N))` -> `e_i`
        COLLAPSE_TUPLE_GET_MAKE_TUPLE = enum.auto()
        #: `tuple_get(i, let(...)(make_tuple()))` -> `let(...)(tuple_get(i, make_tuple()))`
        PROPAGATE_TUPLE_GET = enum.auto()
        #: `{1, 2}` -> `(λ(_tuple_el_1, _tuple_el_2) → {_tuple_el_1, _tuple_el_2})(1, 2)`
        LETIFY_MAKE_TUPLE_ELEMENTS = enum.auto()
        #: `let(tup, {trivial_expr1, trivial_expr2})(foo(tup))`
        #:  -> `foo({trivial_expr1, trivial_expr2})`
        INLINE_TRIVIAL_MAKE_TUPLE = enum.auto()
        #:  Similar as `PROPAGATE_TO_IF_ON_TUPLES`, but propagates in the opposite direction, i.e.
        #:  into the tree, allowing removal of tuple expressions across `if_` calls without
        #:  increasing the size of the tree. This is particularly important for `if` statements
        #:  in the frontend, where outwards propagation can have devastating effects on the tree
        #:  size, without any gained optimization potential. For example
        #:  ```
        #:   complex_lambda(if cond1
        #:     if cond2
        #:       {...}
        #:     else:
        #:       {...}
        #:   else
        #:     {...})
        #:  ```
        #:  is problematic, since `PROPAGATE_TO_IF_ON_TUPLES` would propagate and hence duplicate
        #:  `complex_lambda` three times, while we only want to get rid of the tuple expressions
        #:  inside of the `if_`s.
        #:  Note that this transformation is not mutually exclusive to `PROPAGATE_TO_IF_ON_TUPLES`.
        PROPAGATE_TO_IF_ON_TUPLES_CPS = enum.auto()
        #: `(if cond then {1, 2} else {3, 4})[0]` -> `if cond then {1, 2}[0] else {3, 4}[0]`
        PROPAGATE_TO_IF_ON_TUPLES = enum.auto()
        #: `let((a, let(b, 1)(a_val)))(a)`-> `let(b, 1)(let(a, a_val)(a))`
        PROPAGATE_NESTED_LET = enum.auto()
        #: `let(a, 1)(a)` -> `1` or `let(a, b)(f(a))` -> `f(a)`
        INLINE_TRIVIAL_LET = enum.auto()
        #: `as_fieldop(λ(t) → ·t[0]+·t[1])({a, b})` -> `as_fieldop(λ(a, b) → ·a+·b)(a, b)`
        FLATTEN_AS_FIELDOP_ARGS = enum.auto()
        #: `let(a, b[1])(a)` -> `b[1]`
        INLINE_TRIVIAL_TUPLE_LET_VAR = enum.auto()

        @classmethod
        def all(self) -> CollapseTuple.Transformation:
            return functools.reduce(operator.or_, self.__members__.values())

    uids: eve_utils.UIDGenerator
    ignore_tuple_size: bool
    enabled_transformations: Transformation = Transformation.all()  # noqa: RUF009 [function-call-in-dataclass-default-argument]

    REINFER_TYPES = True

    PRESERVED_ANNEX_ATTRS = ("type", "domain")

    @classmethod
    def apply(
        cls,
        node: itir.Node,
        *,
        ignore_tuple_size: bool = False,
        remove_letified_make_tuple_elements: bool = True,
        offset_provider_type: Optional[common.OffsetProviderType] = None,
        within_stencil: Optional[bool] = None,
        # manually passing enabled transformations is mostly for allowing separate testing of the modes
        enabled_transformations: Optional[Transformation] = None,
        # allow sym references without a symbol declaration, mostly for testing
        allow_undeclared_symbols: bool = False,
        uids: Optional[eve_utils.UIDGenerator] = None,
    ) -> itir.Node:
        """
        Simplifies `make_tuple`, `tuple_get` calls.

        Arguments:
            node: The node to transform.

        Keyword arguments:
            ignore_tuple_size: Apply the transformation even if length of the inner tuple is greater
                than the length of the outer tuple.
            remove_letified_make_tuple_elements: Run `InlineLambdas` as a post-processing step
                to remove left-overs from `LETIFY_MAKE_TUPLE_ELEMENTS` transformation.
                `(λ(_tuple_el_1, _tuple_el_2) → {_tuple_el_1, _tuple_el_2})(1, 2)` -> {1, 2}`
        """
        enabled_transformations = enabled_transformations or cls.enabled_transformations
        offset_provider_type = offset_provider_type or {}
        uids = uids or eve_utils.UIDGenerator()

        if isinstance(node, itir.Program):
            within_stencil = False
        assert within_stencil in [
            True,
            False,
        ], "Parameter 'within_stencil' mandatory if node is not a 'Program'."

        if not ignore_tuple_size:
            node = itir_type_inference.infer(
                node,
                offset_provider_type=offset_provider_type,
                allow_undeclared_symbols=allow_undeclared_symbols,
            )

        new_node = cls(
            ignore_tuple_size=ignore_tuple_size,
            enabled_transformations=enabled_transformations,
            uids=uids,
        ).visit(node, within_stencil=within_stencil)

        # inline to remove left-overs from LETIFY_MAKE_TUPLE_ELEMENTS. this is important
        # as otherwise two equal expressions containing a tuple will not be equal anymore
        # and the CSE pass can not remove them.
        # TODO(tehrengruber): test case for `scan(lambda carry: {1, 2})`
        #  (see solve_nonhydro_stencil_52_like_z_q_tup)
        if remove_letified_make_tuple_elements:
            new_node = inline_lambdas.InlineLambdas.apply(
                new_node, opcount_preserving=True, force_inline_lambda_args=False
            )

        return new_node

    def visit(self, node, **kwargs):
        if cpm.is_call_to(node, "as_fieldop"):
            kwargs = {**kwargs, "within_stencil": True}

        return super().visit(node, **kwargs)

    def transform_collapse_make_tuple_tuple_get(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        if cpm.is_call_to(node, "make_tuple") and all(
            cpm.is_call_to(arg, "tuple_get") for arg in node.args
        ):
            # `make_tuple(tuple_get(0, t), tuple_get(1, t), ..., tuple_get(N-1,t))` -> `t`
            assert len(node.args) > 0 and isinstance(node.args[0], itir.FunCall)
            first_expr = node.args[0].args[1]

            for i, v in enumerate(node.args):
                assert isinstance(v, itir.FunCall)
                assert isinstance(v.args[0], itir.Literal)
                if not (int(v.args[0].value) == i and ir_misc.is_equal(v.args[1], first_expr)):
                    # tuple argument differs, just continue with the rest of the tree
                    return None

            itir_type_inference.reinfer(first_expr)  # type is needed so reinfer on-demand
            assert self.ignore_tuple_size or isinstance(
                first_expr.type, (ts.TupleType, ts.DeferredType)
            )
            if self.ignore_tuple_size or (
                isinstance(first_expr.type, ts.TupleType)
                and len(first_expr.type.types) == len(node.args)
            ):
                return first_expr
        return None

    def transform_collapse_tuple_get_make_tuple(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        if (
            cpm.is_call_to(node, "tuple_get")
            and isinstance(node.args[0], itir.Literal)
            and cpm.is_call_to(node.args[1], "make_tuple")
        ):
            # `tuple_get(i, make_tuple(e_0, e_1, ..., e_i, ..., e_N))` -> `e_i`
            assert not node.args[0].type or type_info.is_integer(node.args[0].type)
            make_tuple_call = node.args[1]
            idx = int(node.args[0].value)
            assert idx < len(make_tuple_call.args), (
                f"Index {idx} is out of bounds for tuple of size {len(make_tuple_call.args)}"
            )
            return node.args[1].args[idx]
        return None

    def transform_propagate_tuple_get(self, node: itir.FunCall, **kwargs) -> Optional[itir.Node]:
        if cpm.is_call_to(node, "tuple_get") and isinstance(node.args[0], itir.Literal):
            # TODO(tehrengruber): extend to general symbols as long as the tail call in the let
            #   does not capture
            # `tuple_get(i, let(...)(make_tuple()))` -> `let(...)(tuple_get(i, make_tuple()))`
            idx, expr = node.args
            assert isinstance(idx, itir.Literal)
            if cpm.is_let(expr):
                return im.call(
                    im.lambda_(*expr.fun.params)(
                        self.fp_transform(im.tuple_get(idx.value, expr.fun.expr), **kwargs)
                    )
                )(*expr.args)
            elif cpm.is_call_to(expr, ("if_", "concat_where")):
                fun = expr.fun
                cond, true_branch, false_branch = expr.args
                return im.call(fun)(
                    cond,
                    self.fp_transform(im.tuple_get(idx.value, true_branch), **kwargs),
                    self.fp_transform(im.tuple_get(idx.value, false_branch), **kwargs),
                )
        return None

    def transform_letify_make_tuple_elements(
        self, node: itir.Node, **kwargs
    ) -> Optional[itir.Node]:
        if cpm.is_call_to(node, "make_tuple"):
            # `make_tuple(expr1, expr1)`
            # -> `let((_tuple_el_1, expr1), (_tuple_el_2, expr2))(make_tuple(_tuple_el_1, _tuple_el_2))`
            bound_vars: dict[itir.Sym, itir.Expr] = {}
            new_args: list[itir.Expr] = []
            for arg in node.args:
                if cpm.is_call_to(node, "make_tuple") and not _is_trivial_make_tuple_call(node):
                    new_arg = im.ref(self.uids.sequential_id(prefix="__ct_el"), arg.type)
                    self._preserve_annex(arg, new_arg)
                    new_args.append(new_arg)
                    bound_vars[im.sym(new_arg.id, arg.type)] = arg
                else:
                    new_args.append(arg)

            if bound_vars:
                return self.fp_transform(
                    im.let(*bound_vars.items())(im.call(node.fun)(*new_args)), **kwargs
                )
        return None

    def transform_inline_trivial_make_tuple(self, node: itir.Node, **kwargs) -> Optional[itir.Node]:
        if cpm.is_let(node):
            # `let(tup, make_tuple(trivial_expr1, trivial_expr2))(foo(tup))`
            #  -> `foo(make_tuple(trivial_expr1, trivial_expr2))`
            eligible_params = [_is_trivial_make_tuple_call(arg) for arg in node.args]
            if any(eligible_params):
                return self.visit(
                    inline_lambdas.inline_lambda(node, eligible_params=eligible_params), **kwargs
                )
        return None

    def transform_propagate_to_if_on_tuples(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        if kwargs["within_stencil"]:
            # TODO(tehrengruber): This significantly increases the size of the tree. Skip transformation
            #  in local-view for now. Revisit.
            return None

        if isinstance(node, itir.FunCall) and not cpm.is_call_to(node, "if_"):
            # TODO(tehrengruber): Only inline if type of branch value is a tuple.
            # Examples:
            # `(if cond then {1, 2} else {3, 4})[0]` -> `if cond then {1, 2}[0] else {3, 4}[0]`
            # `let (b, if cond then {1, 2} else {3, 4})) b[0]`
            #  -> `if cond then let(b, {1, 2})(b[0]) else let(b, {3, 4})(b[0])`
            for i, arg in enumerate(node.args):
                if cpm.is_call_to(arg, "if_"):
                    cond, true_branch, false_branch = arg.args
                    new_true_branch = self.fp_transform(
                        ir_misc.with_altered_arg(node, i, true_branch), **kwargs
                    )
                    new_false_branch = self.fp_transform(
                        ir_misc.with_altered_arg(node, i, false_branch), **kwargs
                    )
                    return im.if_(cond, new_true_branch, new_false_branch)
        return None

    def transform_propagate_to_if_on_tuples_cps(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        # The basic idea of this transformation is to remove tuples across if-stmts by rewriting
        # the expression in continuation passing style, e.g. something like a tuple reordering
        # ```
        # let t = if True then {1, 2} else {3, 4} in
        #   {t[1], t[0]})
        # end
        # ```
        # is rewritten into:
        # ```
        # let cont = λ(el0, el1) → {el1, el0} in
        #  if True then cont(1, 2) else cont(3, 4)
        # end
        # ```
        # Note how the `make_tuple` call argument of the `if` disappears. Since lambda functions
        # are currently inlined (due to limitations of the domain inference) we will only
        # gain something compared `PROPAGATE_TO_IF_ON_TUPLES` if the continuation `cont` is trivial,
        # e.g. a `make_tuple` call like in the example. In that case we can inline the trivial
        # continuation and end up with an only moderately larger tree, e.g.
        # `if True then {2, 1} else {4, 3}`. The examples in the comments below all refer to this
        # tuple reordering example here.

        if not isinstance(node, itir.FunCall) or cpm.is_call_to(node, "if_"):
            return None

        # The first argument that is eligible also transforms all remaining args (They will be
        # part of the continuation which is recursively transformed).
        for i, arg in enumerate(node.args):
            if cpm.is_call_to(arg, "if_"):
                itir_type_inference.reinfer(arg)

                cond, true_branch, false_branch = arg.args  # e.g. `True`, `{1, 2}`, `{3, 4}`
                if not any(
                    isinstance(branch.type, ts.TupleType) for branch in [true_branch, false_branch]
                ):
                    continue
                tuple_type: ts.TupleType = true_branch.type  # type: ignore[assignment]  # type ensured above
                tuple_len = len(tuple_type.types)

                # build and simplify continuation, e.g. λ(el0, el1) → {el1, el0}
                itir_type_inference.reinfer(node)
                assert node.type
                f_type = ts.FunctionType(  # type of continuation in order to keep full type info
                    pos_only_args=tuple_type.types,
                    pos_or_kw_args={},
                    kw_only_args={},
                    returns=node.type,
                )
                f_params = [
                    im.sym(self.uids.sequential_id(prefix="__ct_el_cps"), type_)
                    for type_ in tuple_type.types
                ]
                f_args = [im.ref(param.id, param.type) for param in f_params]
                f_body = ir_misc.with_altered_arg(node, i, im.make_tuple(*f_args))

                # if the function is not trivial the transformation would create a larger tree
                # after inlining so we skip transformation this argument.
                if not _is_trivial_or_tuple_thereof_expr(f_body):
                    continue

                # simplify, e.g., inline trivial make_tuple args
                new_f_body = self.fp_transform(f_body, **kwargs)
                # if the continuation did not simplify there is nothing to gain. Skip
                # transformation of this argument.
                if new_f_body is f_body:
                    continue

                f = im.lambda_(*f_params)(new_f_body)

                # this is the symbol refering to the tuple value inside the two branches of the
                # if, e.g. a symbol refering to `{1, 2}` and `{3, 4}` respectively
                tuple_var = self.uids.sequential_id(prefix="__ct_tuple_cps")
                # this is the symbol refering to our continuation, e.g. `cont` in our example.
                f_var = self.uids.sequential_id(prefix="__ct_cont")
                new_branches = []
                for branch in [true_branch, false_branch]:
                    new_branch = im.let(tuple_var, branch)(
                        im.call(im.ref(f_var, f_type))(  # call to the continuation
                            *(
                                im.tuple_get(i, im.ref(tuple_var, branch.type))
                                for i in range(tuple_len)
                            )
                        )
                    )
                    new_branches.append(self.fp_transform(new_branch, **kwargs))

                # assemble everything together
                new_node = im.let(f_var, f)(im.if_(cond, *new_branches))
                new_node = inline_lambdas.inline_lambda(new_node, eligible_params=[True])
                assert cpm.is_call_to(new_node, "if_")
                new_node = im.if_(
                    cond, *(self.fp_transform(branch, **kwargs) for branch in new_node.args[1:])
                )
                return new_node

        return None

    def transform_propagate_nested_let(self, node: itir.FunCall, **kwargs) -> Optional[itir.Node]:
        if cpm.is_let(node):
            # `let((a, let(b, 1)(a_val)))(a)`-> `let(b, 1)(let(a, a_val)(a))`
            outer_vars: dict[itir.Sym, itir.Expr] = {}
            inner_vars: dict[itir.Sym, itir.Expr] = {}
            original_inner_expr = node.fun.expr
            for arg_sym, arg in zip(node.fun.params, node.args):
                assert arg_sym not in inner_vars
                if cpm.is_let(arg):
                    rename_map: dict[
                        str, ir.SymRef
                    ] = {}  # mapping from symbol with a collision to its new (unique) name
                    for sym, val in zip(arg.fun.params, arg.args):
                        unique_sym = ir_misc.unique_symbol(sym, [s.id for s in outer_vars.keys()])
                        if sym != unique_sym:  # name collision, rename symbol to unique_sym later
                            rename_map[sym.id] = im.ref(unique_sym.id, sym.type)

                        outer_vars[unique_sym] = val

                    new_expr = arg.fun.expr
                    if rename_map:
                        new_expr = inline_lambdas.rename_symbols(new_expr, rename_map)

                    inner_vars[arg_sym] = new_expr
                else:
                    inner_vars[arg_sym] = arg
            if outer_vars:
                return self.fp_transform(
                    im.let(*outer_vars.items())(
                        self.fp_transform(
                            im.let(*inner_vars.items())(original_inner_expr), **kwargs
                        )
                    ),
                    **kwargs,
                )
        return None

    def transform_inline_trivial_let(self, node: itir.FunCall, **kwargs) -> Optional[itir.Node]:
        if cpm.is_let(node):
            if isinstance(node.fun.expr, itir.SymRef):
                # `let(a, 1)(a)` -> `1`
                for arg_sym, arg in zip(node.fun.params, node.args):
                    if isinstance(node.fun.expr, itir.SymRef) and node.fun.expr.id == arg_sym.id:
                        return arg
            if any(
                trivial_args := [isinstance(arg, (itir.SymRef, itir.Literal)) for arg in node.args]
            ):
                return inline_lambdas.inline_lambda(node, eligible_params=trivial_args)

        return None

    def transform_inline_trivial_tuple_let_var(self, node: ir.Node, **kwargs) -> Optional[ir.Node]:
        if cpm.is_let(node):
            if any(trivial_args := [_is_trivial_or_tuple_thereof_expr(arg) for arg in node.args]):
                return inline_lambdas.inline_lambda(node, eligible_params=trivial_args)
        return None

    # TODO(tehrengruber): This is a transformation that should be executed before visiting the children. Then
    #  revisiting the body would not be needed.
    def transform_flatten_as_fieldop_args(
        self, node: itir.FunCall, **kwargs
    ) -> Optional[itir.Node]:
        # `as_fieldop(λ(t) → ·t[0]+·t[1])({a, b})` -> `as_fieldop(λ(a, b) → ·a+·b)(a, b)`
        if not cpm.is_applied_as_fieldop(node):
            return None

        for arg in node.args:
            itir_type_inference.reinfer(arg)

        if not any(isinstance(arg.type, ts.TupleType) for arg in node.args):
            return None

        node = ir_misc.canonicalize_as_fieldop(node)
        stencil, restore_scan = ir_misc.unwrap_scan(
            node.fun.args[0]  # type: ignore[attr-defined] # ensured by cpm.is_applied_as_fieldop
        )

        new_body = stencil.expr
        domain = node.fun.args[1] if len(node.fun.args) > 1 else None  # type: ignore[attr-defined] # ensured by cpm.is_applied_as_fieldop
        remapped_args: dict[
            itir.Sym, itir.Expr
        ] = {}  # contains the arguments that are remapped, e.g. `{a, b}`
        new_params: list[itir.Sym] = []
        new_args: list[itir.Expr] = []
        for param, arg in zip(stencil.params, node.args, strict=True):
            if isinstance(arg.type, ts.TupleType):
                assert isinstance(param.type, it_ts.IteratorType)
                ref_to_remapped_arg = im.ref(
                    f"__ct_flat_remapped_{len(remapped_args)}",
                    arg.type,
                )
                self._preserve_annex(arg, ref_to_remapped_arg)
                remapped_args[im.sym(ref_to_remapped_arg.id, arg.type)] = arg
                new_params_inner, lift_params = [], []
                assert isinstance(param.type.element_type, ts.TupleType)
                for i, type_ in enumerate(param.type.element_type.types):
                    assert isinstance(type_, ts.DataType)
                    new_param = im.sym(
                        _flattened_as_fieldop_param_el_name(param.id, i),
                        _with_altered_iterator_element_type(param.type, type_),
                    )
                    lift_params.append(
                        im.sym(
                            new_param.id,
                            _with_altered_iterator_position_dims(new_param.type, "unknown"),  # type: ignore[arg-type]  # always in IteratorType
                        )
                    )
                    new_params_inner.append(new_param)
                    new_args.append(im.tuple_get(i, ref_to_remapped_arg))

                # an iterator that substitutes the original (tuple) iterator, e.g. `t`. Built
                # from the new parameters which are the elements of `t`.
                param_substitute = im.lift(
                    im.lambda_(*lift_params)(
                        im.make_tuple(*[im.deref(im.ref(p.id, p.type)) for p in lift_params])
                    )
                )(*[im.ref(p.id, p.type) for p in new_params_inner])

                new_body = im.let(param.id, param_substitute)(new_body)
                # note: the lift is trivial so inlining it is not an issue with respect to tree size
                new_body = inline_lambdas.inline_lambda(new_body, force_inline_lift_args=True)

                new_params.extend(new_params_inner)
            else:
                new_params.append(param)
                new_args.append(arg)

        # remove lifts again
        new_body = inline_lifts.InlineLifts(
            flags=inline_lifts.InlineLifts.Flag.INLINE_DEREF_LIFT
            | inline_lifts.InlineLifts.Flag.PROPAGATE_SHIFT
        ).visit(new_body)
        new_body = self.visit(new_body, **kwargs)
        new_stencil = restore_scan(im.lambda_(*new_params)(new_body))

        return im.let(*remapped_args.items())(im.as_fieldop(new_stencil, domain)(*new_args))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
import dataclasses

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
)


@dataclasses.dataclass
class RemoveBroadcast(PreserveLocationVisitor, NodeTranslator):
    """
    Transforms calls to 'broadcast' into calls to 'as_fieldop' with 'deref'
    and the respective domain from node.annex.

    Example:
    >>> from gt4py.next import Dimension, common
    >>> IDim = Dimension("IDim")
    >>> JDim = Dimension("JDim")
    >>> domain = im.domain(common.GridType.CARTESIAN, {IDim: (0, 10), JDim: (0, 10)})
    >>> expr = im.call("broadcast")(
    ...     im.ref("inp"),
    ...     im.make_tuple(
    ...         *(itir.AxisLiteral(value=dim.value, kind=dim.kind) for dim in (IDim, JDim))
    ...     ),
    ... )
    >>> expr.annex.domain = domain_utils.SymbolicDomain.from_expr(domain)
    >>> transformed = RemoveBroadcast.apply(expr)
    >>> print(transformed)
    as_fieldop(deref, c⟨ IDimₕ: [0, 10[, JDimₕ: [0, 10[ ⟩)(inp)
    """

    PRESERVED_ANNEX_ATTRS = ("domain",)

    @classmethod
    def apply(cls, node: itir.Node):
        return cls().visit(node)

    def visit_FunCall(self, node: itir.FunCall) -> itir.FunCall:
        node = self.generic_visit(node)

        if cpm.is_call_to(node, "broadcast"):
            expr = node.args[0]
            node = im.as_fieldop("deref", domain_utils.SymbolicDomain.as_expr(node.annex.domain))(
                expr
            )
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import pathlib
import tempfile
from typing import Any, Optional

import diskcache
import factory
import numpy as np

import gt4py._core.definitions as core_defs
import gt4py.next.allocators as next_allocators
from gt4py._core import locking
from gt4py.next import backend, common, config, field_utils, metrics
from gt4py.next.embedded import nd_array_field
from gt4py.next.otf import arguments, recipes, stages, workflow
from gt4py.next.otf.binding import nanobind
from gt4py.next.otf.compilation import compiler
from gt4py.next.otf.compilation.build_systems import compiledb
from gt4py.next.program_processors.codegens.gtfn import gtfn_module


def convert_arg(arg: Any) -> Any:
    # Note: this function is on the hot path and needs to have minimal overhead.
    if (origin := getattr(arg, "__gt_origin__", None)) is not None:
        # `Field` is the most likely case, we use `__gt_origin__` as the property is needed anyway
        # and (currently) uniquely identifies a `NDArrayField` (which is the only supported `Field`)
        assert isinstance(arg, nd_array_field.NdArrayField)
        return arg.ndarray, origin
    if isinstance(arg, tuple):
        return tuple(convert_arg(a) for a in arg)
    if isinstance(arg, np.bool_):
        # nanobind does not support implicit conversion of `np.bool` to `bool`
        return bool(arg)
    # TODO(havogt): if this function still appears in profiles,
    # we should avoid going through the previous isinstance checks for detecting a scalar.
    # E.g. functools.cache on the arg type, returning a function that does the conversion
    return arg


def convert_args(
    inp: stages.ExtendedCompiledProgram, device: core_defs.DeviceType = core_defs.DeviceType.CPU
) -> stages.CompiledProgram:
    def decorated_program(
        *args: Any,
        offset_provider: dict[str, common.Connectivity | common.Dimension],
        out: Any = None,
    ) -> None:
        # Note: this function is on the hot path and needs to have minimal overhead.
        if out is not None:
            args = (*args, out)
        converted_args = (convert_arg(arg) for arg in args)
        conn_args = extract_connectivity_args(offset_provider, device)

        opt_kwargs: dict[str, Any]
        metric_collection = metrics.get_active_metric_collection()
        if collect_metrics := (
            metric_collection is not None and (config.COLLECT_METRICS_LEVEL >= metrics.PERFORMANCE)
        ):
            # If we are collecting metrics, we need to add the `exec_info` argument
            # to the `inp` call, which will be used to collect performance metrics.
            exec_info: dict[str, float] = {}
            opt_kwargs = {"exec_info": exec_info}
        else:
            opt_kwargs = {}

        # generate implicit domain size arguments only if necessary, using `iter_size_args()`
        inp(
            *converted_args,
            *(arguments.iter_size_args(args) if inp.implicit_domain else ()),
            *conn_args,
            **opt_kwargs,
        )

        if collect_metrics:
            assert metric_collection is not None
            value = exec_info["run_cpp_end_time"] - exec_info["run_cpp_start_time"]
            metric_collection.add_sample(metrics.COMPUTE_METRIC, value)

    return decorated_program


def extract_connectivity_args(
    offset_provider: dict[str, common.Connectivity | common.Dimension], device: core_defs.DeviceType
) -> list[tuple[core_defs.NDArrayObject, tuple[int, ...]]]:
    # Note: this function is on the hot path and needs to have minimal overhead.
    args: list[tuple[core_defs.NDArrayObject, tuple[int, ...]]] = []
    # Note: the order here needs to agree with the order of the generated bindings
    for conn in offset_provider.values():
        if (ndarray := getattr(conn, "ndarray", None)) is not None:
            assert common.is_neighbor_table(conn)
            assert field_utils.verify_device_field_type(conn, device)
            args.append((ndarray, (0, 0)))
            continue
        assert isinstance(conn, common.Dimension)
    return args


class FileCache(diskcache.Cache):
    """
    This class extends `diskcache.Cache` to ensure the cache is properly
    - opened when accessed by multiple processes using a file lock. This guards the creating of the
    cache object, which has been reported to cause `sqlite3.OperationalError: database is locked`
    errors and slow startup times when multiple processes access the cache concurrently. While this
    issue occurred frequently and was observed to be fixed on distributed file systems, the lock
    does not guarantee correct behavior in particular for accesses to the cache (beyond opening)
    since the underlying SQLite database is unreliable when stored on an NFS based file system.
    It does however ensure correctness of concurrent cache accesses on a local file system. See
    #1745 for more details.
    - closed upon deletion, i.e. it ensures that any resources associated with the cache are
    properly released when the instance is garbage collected.
    """

    def __init__(self, directory: Optional[str | pathlib.Path] = None, **settings: Any) -> None:
        if directory:
            lock_dir = pathlib.Path(directory).parent
        else:
            lock_dir = pathlib.Path(tempfile.gettempdir())

        lock_dir.mkdir(parents=True, exist_ok=True)
        with locking.lock(lock_dir):
            super().__init__(directory=directory, **settings)

        self._init_complete = True

    def __del__(self) -> None:
        if getattr(self, "_init_complete", False):  # skip if `__init__` didn't finished
            self.close()


class GTFNCompileWorkflowFactory(factory.Factory):
    class Meta:
        model = recipes.OTFCompileWorkflow

    class Params:
        device_type: core_defs.DeviceType = core_defs.DeviceType.CPU
        cmake_build_type: config.CMakeBuildType = factory.LazyFunction(
            lambda: config.CMAKE_BUILD_TYPE
        )
        builder_factory: compiler.BuildSystemProjectGenerator = factory.LazyAttribute(
            lambda o: compiledb.CompiledbFactory(cmake_build_type=o.cmake_build_type)
        )

        cached_translation = factory.Trait(
            translation=factory.LazyAttribute(
                lambda o: workflow.CachedStep(
                    o.bare_translation,
                    hash_function=stages.fingerprint_compilable_program,
                    cache=FileCache(str(config.BUILD_CACHE_DIR / "gtfn_cache")),
                )
            ),
        )

        bare_translation = factory.SubFactory(
            gtfn_module.GTFNTranslationStepFactory,
            device_type=factory.SelfAttribute("..device_type"),
        )

    translation = factory.LazyAttribute(lambda o: o.bare_translation)

    bindings: workflow.Workflow[stages.ProgramSource, stages.CompilableSource] = (
        nanobind.bind_source
    )
    compilation = factory.SubFactory(
        compiler.CompilerFactory,
        cache_lifetime=factory.LazyFunction(lambda: config.BUILD_CACHE_LIFETIME),
        builder_factory=factory.SelfAttribute("..builder_factory"),
    )
    decoration = factory.LazyAttribute(
        lambda o: functools.partial(convert_args, device=o.device_type)
    )


class GTFNBackendFactory(factory.Factory):
    class Meta:
        model = backend.Backend

    class Params:
        name_device = "cpu"
        name_cached = ""
        name_temps = ""
        name_postfix = ""
        gpu = factory.Trait(
            allocator=next_allocators.StandardGPUFieldBufferAllocator(),
            device_type=core_defs.CUPY_DEVICE_TYPE or core_defs.DeviceType.CUDA,
            name_device="gpu",
        )
        cached = factory.Trait(
            executor=factory.LazyAttribute(
                lambda o: workflow.CachedStep(o.otf_workflow, hash_function=o.hash_function)
            ),
            name_cached="_cached",
        )
        device_type = core_defs.DeviceType.CPU
        hash_function = stages.compilation_hash
        otf_workflow = factory.SubFactory(
            GTFNCompileWorkflowFactory, device_type=factory.SelfAttribute("..device_type")
        )

    name = factory.LazyAttribute(
        lambda o: f"run_gtfn_{o.name_device}{o.name_temps}{o.name_cached}{o.name_postfix}"
    )

    executor = factory.LazyAttribute(lambda o: o.otf_workflow)
    allocator = next_allocators.StandardCPUFieldBufferAllocator()
    transforms = backend.DEFAULT_TRANSFORMS


run_gtfn = GTFNBackendFactory()

run_gtfn_imperative = GTFNBackendFactory(
    name_postfix="_imperative", otf_workflow__translation__use_imperative_backend=True
)

run_gtfn_cached = GTFNBackendFactory(cached=True, otf_workflow__cached_translation=True)

run_gtfn_gpu = GTFNBackendFactory(gpu=True)

run_gtfn_gpu_cached = GTFNBackendFactory(
    gpu=True, cached=True, otf_workflow__cached_translation=True
)

run_gtfn_no_transforms = GTFNBackendFactory(
    otf_workflow__bare_translation__enable_itir_transforms=False
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import collections
import dataclasses
import math
from typing import Callable, Iterable, TypeVar, Union, cast

import gt4py.next.iterator.ir_utils.ir_makers as im
from gt4py.eve import (
    NodeTranslator,
    NodeVisitor,
    PreserveLocationVisitor,
    SymbolTableTrait,
    VisitorWithSymbolTableTrait,
)
from gt4py.eve.utils import UIDGenerator
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm
from gt4py.next.iterator.transforms.inline_lambdas import inline_lambda
from gt4py.next.iterator.type_system import inference as itir_type_inference
from gt4py.next.type_system import type_info, type_specifications as ts


def _is_trivial_tuple_expr(node: itir.Expr):
    """Return if node is a `make_tuple` call with all elements `SymRef`s, `Literal`s or tuples thereof."""
    if cpm.is_call_to(node, "make_tuple") and all(
        isinstance(arg, (itir.SymRef, itir.Literal)) or _is_trivial_tuple_expr(arg)
        for arg in node.args
    ):
        return True
    if cpm.is_call_to(node, "tuple_get") and (
        isinstance(node.args[1], (itir.SymRef, itir.Literal))
        or _is_trivial_tuple_expr(node.args[1])
    ):
        return True
    return False


@dataclasses.dataclass
class _NodeReplacer(PreserveLocationVisitor, NodeTranslator):
    PRESERVED_ANNEX_ATTRS = ("type", "domain")

    expr_map: dict[int, itir.SymRef]

    def visit_Expr(self, node: itir.Node) -> itir.Node:
        if id(node) in self.expr_map:
            return self.expr_map[id(node)]
        return self.generic_visit(node)

    def visit_FunCall(self, node: itir.FunCall) -> itir.Node:
        node = cast(itir.FunCall, self.visit_Expr(node))
        # TODO(tehrengruber): Use symbol name from the inner let, to increase readability of IR
        # If we encounter an expression like:
        #  (λ(_cs_1) → (λ(a) → a+a)(_cs_1))(outer_expr)
        # (non-recursively) inline the lambda to obtain:
        #  (λ(_cs_1) → _cs_1+_cs_1)(outer_expr)
        # In the CSE this allows identifying more common subexpressions later on. Other users
        # of `extract_subexpression` (e.g. temporary extraction) can also rely on this to avoid
        # the need to handle this artificial let-statements.
        if isinstance(node, itir.FunCall) and isinstance(node.fun, itir.Lambda):
            eligible_params = [isinstance(arg, itir.SymRef) for arg in node.args]
            if any(eligible_params):
                # note: the inline is opcount preserving anyway so avoid the additional
                # effort in the inliner by disabling opcount preservation.
                return inline_lambda(
                    node, opcount_preserving=False, eligible_params=eligible_params
                )
        return node


def _is_collectable_expr(node: itir.Node) -> bool:
    if isinstance(node, itir.FunCall):
        # do not collect (and thus deduplicate in CSE) shift(offsets…) calls. Node must still be
        #  visited, to ensure symbol dependencies are recognized correctly.
        # do also not collect reduce nodes if they are left in the IR at this point, this may lead to
        #  conceptual problems (other parts of the tool chain rely on the arguments being present directly
        #  on the reduce FunCall node (connectivity deduction)), as well as problems with the imperative backend
        #  backend (single pass eager depth first visit approach)
        # do also not collect lifts or applied lifts as they become invisible to the lift inliner
        #  otherwise
        # do also not collect index nodes because otherwise the right hand side of SetAts becomes a let statement
        #  instead of an as_fieldop
        if cpm.is_call_to(
            node, ("lift", "shift", "reduce", "map_", "index")
        ) or cpm.is_applied_lift(node):
            return False
        return True
    # do also not collect make_tuple(index) nodes because otherwise the right hand side of SetAts becomes a let statement
    #  instead of an as_fieldop
    if cpm.is_call_to(node, "make_tuple") and all(
        cpm.is_call_to(arg, "index") for arg in node.args
    ):
        return False
    elif isinstance(node, itir.Lambda):
        return True

    return False


@dataclasses.dataclass
class CollectSubexpressions(PreserveLocationVisitor, VisitorWithSymbolTableTrait, NodeVisitor):
    @dataclasses.dataclass
    class SubexpressionData:
        #: A list of node ids with equal hash and a set of collected child subexpression ids
        subexprs: list[tuple[int, set[int]]] = dataclasses.field(default_factory=list)
        #: Maximum depth of a subexpression in the tree. Used to sort collected subexpressions
        #:  such that deeper nodes can be processed (in other passed building upon this pass)
        #:  earlier.
        max_depth: int | float = -math.inf

    @dataclasses.dataclass
    class State:
        #: A dictionary mapping a node to a list of node ids with equal hash and some additional
        #: information. See `SubexpressionData` for more information.
        subexprs: dict[itir.Node, CollectSubexpressions.SubexpressionData] = dataclasses.field(
            default_factory=dict
        )
        # TODO(tehrengruber): Revisit if this makes sense or if we can just recompute the collected
        #  child node ids and get simpler code.
        #: The ids of all child subexpressions which are collected.
        collected_child_node_ids: set[int] = dataclasses.field(default_factory=set)
        #: The ids of all nodes declaring a symbol which are referenced (using a `SymRef`)
        used_symbol_ids: set[int] = dataclasses.field(default_factory=set)

        def remove_subexprs(self, nodes: Iterable[itir.Node]) -> None:
            node_ids_to_remove: set[int] = set()
            for node in nodes:
                subexpr_data = self.subexprs.pop(node, None)
                if subexpr_data:
                    node_ids, _ = zip(*subexpr_data.subexprs)
                    node_ids_to_remove |= set(node_ids)
            for subexpr_data in self.subexprs.values():
                for _, collected_child_node_ids in subexpr_data.subexprs:
                    collected_child_node_ids -= node_ids_to_remove

    @classmethod
    def apply(cls, node: itir.Node) -> dict[itir.Node, list[tuple[int, set[int]]]]:
        state = cls.State()
        obj = cls()
        obj.visit(node, state=state, depth=-1)
        # Return subexpression such that the nodes closer to the root come first and skip the root
        # node itself.
        subexprs_sorted: list[tuple[itir.Node, CollectSubexpressions.SubexpressionData]] = sorted(
            state.subexprs.items(), key=lambda el: el[1].max_depth
        )
        return {k: v.subexprs for k, v in subexprs_sorted if k is not node}

    def generic_visit(self, node, **kwargs):
        depth = kwargs.pop("depth")
        return super().generic_visit(node, depth=depth + 1, **kwargs)

    def visit(self, node: itir.Node, is_let_form: bool = False, **kwargs) -> None:  # type: ignore[override] # supertype accepts any node, but we want to be more specific here.
        # do not descent into un-applied lambda
        can_collect_children = not isinstance(node, itir.Lambda) or is_let_form

        if (
            can_collect_children
            and not isinstance(node, SymbolTableTrait)
            and not _is_collectable_expr(node)
        ):
            return super().visit(node, **kwargs)

        depth = kwargs["depth"]
        parent_state = kwargs.pop("state")
        collected_child_node_ids: set[int] = set()
        used_symbol_ids: set[int] = set()

        if can_collect_children:
            # Special handling of `if_(condition, true_branch, false_branch)` like expressions that
            # avoids extracting subexpressions unless they are used in either the condition or both
            # branches.
            if cpm.is_call_to(node, "if_"):
                assert len(node.args) == 3
                # collect subexpressions for all arguments to the `if_`
                arg_states = [self.State() for _ in node.args]
                for arg, state in zip(node.args, arg_states):
                    self.visit(arg, state=state, **(kwargs | {"depth": depth + 1}))

                # remove all subexpressions that are not eligible for collection
                #  (either they occur in the condition or in both branches)
                eligible_subexprs = arg_states[0].subexprs.keys() | (
                    arg_states[1].subexprs.keys() & arg_states[2].subexprs.keys()
                )
                for arg_state in arg_states:
                    arg_state.remove_subexprs(arg_state.subexprs.keys() - eligible_subexprs)

                # merge the states of the three arguments
                subexprs: dict[itir.Node, CollectSubexpressions.SubexpressionData] = (
                    collections.defaultdict(self.SubexpressionData)
                )
                for state in arg_states:
                    for subexpr, data in state.subexprs.items():
                        merged_data = subexprs[subexpr]
                        merged_data.subexprs.extend(data.subexprs)
                        merged_data.max_depth = max(merged_data.max_depth, data.max_depth)
                collected_child_node_ids = set.union(
                    *(state.collected_child_node_ids for state in arg_states)
                )
                used_symbol_ids = set.union(*(state.used_symbol_ids for state in arg_states))
                # propagate collected subexpressions to parent
                for subexpr, data in subexprs.items():
                    parent_data = parent_state.subexprs.setdefault(
                        subexpr, self.SubexpressionData()
                    )
                    parent_data.subexprs.extend(data.subexprs)
                    parent_data.max_depth = max(merged_data.max_depth, data.max_depth)
            elif cpm.is_let(node):
                new_state = self.State(
                    parent_state.subexprs, collected_child_node_ids, used_symbol_ids
                )
                self.visit(
                    node.fun, state=new_state, is_let_form=True, **(kwargs | {"depth": depth + 1})
                )
                for arg in node.args:
                    self.visit(arg, state=new_state, **(kwargs | {"depth": depth + 1}))
            else:
                super().visit(
                    node,
                    state=self.State(
                        parent_state.subexprs, collected_child_node_ids, used_symbol_ids
                    ),
                    **kwargs,
                )
        else:
            # we don't want to collect any expression, but still need to track all symbol refs
            super().visit(node, state=self.State({}, set(), used_symbol_ids), **kwargs)

        if isinstance(node, SymbolTableTrait):
            # remove symbols used in child nodes if they are declared in the current node
            used_symbol_ids = used_symbol_ids - {id(v) for v in node.annex.symtable.values()}

        # if no symbols are used that are defined in the root node, i.e. the node given to `apply`,
        # we collect the subexpression
        if not used_symbol_ids and _is_collectable_expr(node):
            parent_data = parent_state.subexprs.setdefault(node, self.SubexpressionData())
            parent_data.subexprs.append((id(node), collected_child_node_ids))
            parent_data.max_depth = max(parent_data.max_depth, depth)

            # propagate to parent that we have collected its child
            parent_state.collected_child_node_ids.add(id(node))

        # propagate used symbol ids to parent
        parent_state.used_symbol_ids.update(used_symbol_ids)

        # propagate to parent which of its children we have collected
        # TODO(tehrengruber): This is expensive for a large tree. Use something like a "ChainSet".
        parent_state.collected_child_node_ids.update(collected_child_node_ids)

    def visit_SymRef(
        self, node: itir.SymRef, *, symtable: dict[str, itir.Node], state: State, **kwargs
    ) -> None:
        if node.id in symtable:  # root symbol otherwise
            state.used_symbol_ids.add(id(symtable[node.id]))


def extract_subexpression(
    node: itir.Expr,
    predicate: Callable[[itir.Expr, int], bool],
    uid_generator: UIDGenerator,
    once_only: bool = False,
    deepest_expr_first: bool = False,
) -> tuple[itir.Expr, Union[dict[itir.Sym, itir.Expr], None], bool]:
    """
    Given an expression extract all subexprs and return a new expr with the subexprs replaced.

    The return value is a triplet of
    - the new expr with all extracted subexpressions replaced by a reference to a new symbol
    - a dictionary mapping each new symbol to the respective subexpr that was extracted
    - a boolean indicating if a subexression was not collected because its parent was already
      collected.

     Arguments:
        node: The node to extract from.
        predicate: If this predicate evaluates to true the respective subexpression is extracted.
          Takes a subexpression and the number of occurences of the subexpression in the root node
          as arguments.
        uid_generator: The uid generator used to generate new symbol names.
        once_only: If set extraction is stopped after the first expression that is extracted.
        deepest_expr_first: Extract subexpressions that are lower in the tree first. Otherwise,
          expressions closer to the root are extracted first. Requires `once_only == True`.


    Examples:
        Default case for `(x+y) + ((x+y)+z)`:

        >>> from gt4py.eve.utils import UIDGenerator
        >>> expr = im.plus(im.plus("x", "y"), im.plus(im.plus("x", "y"), "z"))
        >>> predicate = lambda subexpr, num_occurences: num_occurences > 1
        >>> new_expr, extracted_subexprs, _ = extract_subexpression(
        ...     expr, predicate, UIDGenerator(prefix="_subexpr")
        ... )
        >>> print(new_expr)
        _subexpr_1 + (_subexpr_1 + z)
        >>> for sym, subexpr in extracted_subexprs.items():
        ...     print(f"`{sym}`: `{subexpr}`")
        `_subexpr_1`: `x + y`

        The order of the extraction can be configured using `deepest_expr_first`. By default, the nodes
        closer to the root are eliminated first:

        >>> expr = im.plus(
        ...     im.plus(im.plus("x", "y"), im.plus("x", "y")),
        ...     im.plus(im.plus("x", "y"), im.plus("x", "y")),
        ... )
        >>> new_expr, extracted_subexprs, ignored_children = extract_subexpression(
        ...     expr, predicate, UIDGenerator(prefix="_subexpr"), deepest_expr_first=False
        ... )
        >>> print(new_expr)
        _subexpr_1 + _subexpr_1
        >>> for sym, subexpr in extracted_subexprs.items():
        ...     print(f"`{sym}`: `{subexpr}`")
        `_subexpr_1`: `x + y + (x + y)`

        Since `(x+y)` is a child of one of the expressions it is ignored:

        >>> print(ignored_children)
        True

        Setting `deepest_expr_first` will extract nodes deeper in the tree first:

        >>> expr = im.plus(
        ...     im.plus(im.plus("x", "y"), im.plus("x", "y")),
        ...     im.plus(im.plus("x", "y"), im.plus("x", "y")),
        ... )
        >>> new_expr, extracted_subexprs, _ = extract_subexpression(
        ...     expr,
        ...     predicate,
        ...     UIDGenerator(prefix="_subexpr"),
        ...     once_only=True,
        ...     deepest_expr_first=True,
        ... )
        >>> print(new_expr)
        _subexpr_1 + _subexpr_1 + (_subexpr_1 + _subexpr_1)
        >>> for sym, subexpr in extracted_subexprs.items():
        ...     print(f"`{sym}`: `{subexpr}`")
        `_subexpr_1`: `x + y`

        Note that this requires `once_only` to be set right now.
    """
    if deepest_expr_first and not once_only:
        # TODO(tehrengruber): Revisit. We could fix this, but is this case even needed?
        # If we traverse the deepest nodes first, but don't stop after the first extraction the new
        # expression is not meaningful in the current implementation. Just disallow this case for
        # now. E.g.:
        # Input expression:
        #   `((x + y) + (x + y)) + ((x + y) + (x + y))`
        # New expression:
        #   `_subexpr_2 + _subexpr_2`
        # Extracted subexpression:
        #  `_subexpr_1`: `x + y`  (This subexpression is not used anywhere)
        #  `_subexpr_2`: `x + y + (x + y)`
        raise NotImplementedError(
            "Results of the current implementation not meaningful for "
            "'deepest_expr_first == True' and 'once_only == True'."
        )

    ignored_children = False
    extracted = dict[itir.Sym, itir.Expr]()

    # collect expressions
    subexprs = CollectSubexpressions.apply(node)

    # collect multiple occurrences and map them to fresh symbols
    expr_map: dict[int, itir.SymRef] = {}
    ignored_ids = set()
    for expr, subexpr_entry in (
        subexprs.items() if not deepest_expr_first else reversed(subexprs.items())
    ):
        # just to make mypy happy when calling the predicate. Every subnode and hence subexpression
        # is an expr anyway.
        assert isinstance(expr, itir.Expr)

        if not predicate(expr, len(subexpr_entry)):
            continue

        eligible_ids = set()
        for id_, child_ids in subexpr_entry:
            if id_ in ignored_ids:
                ignored_children = True
            else:
                eligible_ids.add(id_)
                # since the node id is eligible don't eliminate its children
                ignored_ids.update(child_ids)

        # if no node ids are eligible, e.g. because the parent was already eliminated or because
        # the expression occurs only once, skip the elimination
        if not eligible_ids:
            continue

        expr_id = uid_generator.sequential_id()
        extracted[itir.Sym(id=expr_id)] = expr
        expr_ref = itir.SymRef(id=expr_id)
        for id_ in eligible_ids:
            expr_map[id_] = expr_ref

        if once_only:
            break

    if not expr_map:
        return node, None, False

    return _NodeReplacer(expr_map).visit(node), extracted, ignored_children


ProgramOrExpr = TypeVar("ProgramOrExpr", bound=itir.Program | itir.Expr)


@dataclasses.dataclass(frozen=True)
class CommonSubexpressionElimination(PreserveLocationVisitor, NodeTranslator):
    """
    Perform common subexpression elimination.

    Examples:
        >>> x = itir.SymRef(id="x")
        >>> plus = lambda a, b: itir.FunCall(fun=itir.SymRef(id=("plus")), args=[a, b])
        >>> expr = plus(plus(x, x), plus(x, x))
        >>> print(CommonSubexpressionElimination.apply(expr, within_stencil=True))
        (λ(_cs_1) → _cs_1 + _cs_1)(x + x)

    The pass visits the tree top-down starting from the root node, e.g. an itir.Program.
    For each node we extract (eligible) subexpressions occuring more than once using
    :ref:`extract_subexpression`. In field-view context we only extract expression when they are
    fields (or composites thereof), in local view everything is eligible. Since the visit is
    top-down, extracted expressions always land up in the outermost scope they can appear in.
    """

    # we use one UID generator per instance such that the generated ids are
    # stable across multiple runs (required for caching to properly work)
    uids: UIDGenerator = dataclasses.field(
        init=False, repr=False, default_factory=lambda: UIDGenerator(prefix="_cs")
    )

    collect_all: bool = dataclasses.field(default=False)

    @classmethod
    def apply(
        cls,
        node: ProgramOrExpr,
        within_stencil: bool | None = None,
        offset_provider_type: common.OffsetProviderType | None = None,
    ) -> ProgramOrExpr:
        is_program = isinstance(node, itir.Program)
        if is_program:
            assert within_stencil is None
            within_stencil = False
        else:
            assert within_stencil is not None, (
                "The expression's context must be specified using `within_stencil`."
            )

        offset_provider_type = offset_provider_type or {}
        node = itir_type_inference.infer(
            node, offset_provider_type=offset_provider_type, allow_undeclared_symbols=not is_program
        )
        return cls().visit(node, within_stencil=within_stencil)

    def generic_visit(self, node, **kwargs):
        if cpm.is_call_to(node, "as_fieldop"):
            assert not kwargs.get("within_stencil")
        within_stencil = cpm.is_call_to(node, "as_fieldop") or kwargs.get("within_stencil")

        return super().generic_visit(node, **(kwargs | {"within_stencil": within_stencil}))

    def visit_FunCall(self, node: itir.FunCall, **kwargs):
        within_stencil = kwargs["within_stencil"]

        if cpm.is_call_to(node, ("cartesian_domain", "unstructured_domain")):
            return node

        def predicate(subexpr: itir.Expr, num_occurences: int):
            # note: be careful here with the syntatic context: the expression might be in local
            #  view, even though the syntactic context of `node` is in field view.
            # note: what is extracted is sketched in the docstring above. keep it updated.
            if num_occurences > 1:
                if within_stencil:
                    # TODO(tehrengruber): Lists must not be extracted to avoid errors in partial
                    #  shift detection of UnrollReduce pass. Solve there. See #1795.
                    if isinstance(subexpr.type, ts.ListType):
                        return False
                    return True
                # condition is only necessary since typing on lambdas is not preserved during
                #  the transformation
                elif not isinstance(subexpr, itir.Lambda):
                    # only extract fields outside of `as_fieldop`
                    # `as_fieldop(...)(field_expr, field_expr)`
                    # -> `(λ(_cs_1) → as_fieldop(...)(_cs_1, _cs_1))(field_expr)`
                    # only extract if subexpression is not a trivial tuple expressions, e.g.,
                    # `make_tuple(a, b)`, as this would result in a more costly temporary.
                    assert isinstance(subexpr.type, ts.TypeSpec)
                    if all(
                        isinstance(stype, ts.FieldType)
                        for stype in type_info.primitive_constituents(subexpr.type)
                    ) and not _is_trivial_tuple_expr(subexpr):
                        return True
            return False

        new_expr, extracted, ignored_children = extract_subexpression(node, predicate, self.uids)

        if not extracted:
            return self.generic_visit(node, **kwargs)

        # apply remapping
        result = im.let(*extracted.items())(new_expr)
        itir_type_inference.copy_type(from_=node, to=result, allow_untyped=True)

        # if the node id is ignored (because its parent is eliminated), but it occurs
        # multiple times then we want to visit the final result once more.
        # TODO(tehrengruber): Instead of revisiting we might be able to replace subexpressions
        #  inside of subexpressions directly. This would require a different order of replacement
        #  (from lower to higher level).
        if ignored_children:
            return self.visit(result, **kwargs)

        return self.generic_visit(result, **kwargs)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Optional, Union

import dace
from dace import properties as dace_properties, transformation as dace_transformation
from dace.sdfg import nodes as dace_nodes

from gt4py.next import common as gtx_common
from gt4py.next.program_processors.runners.dace import gtir_to_sdfg_utils


def gt_set_iteration_order(
    sdfg: dace.SDFG,
    unit_strides_dim: Optional[
        Union[str, gtx_common.Dimension, list[Union[str, gtx_common.Dimension]]]
    ] = None,
    unit_strides_kind: Optional[gtx_common.DimensionKind] = None,
    validate: bool = True,
    validate_all: bool = False,
) -> Any:
    """Set the iteration order of the Maps correctly.

    The iteration order, while irrelevant from an SDFG point of view, is highly
    relevant in code and the fastest varying index ("inner most loop" in CPU or
    "x block dimension" in GPU) should be associated with the stride 1 dimension
    of the array.
    This function will reorder the map indexes such that this is the case.

    See the `MapIterationOrder` transformation for more.

    Args:
        sdfg: The SDFG to process.
        unit_strides_dims: List of dimensions that is considered to have unit strides.
        unit_strides_kind: All dimensions of this kind are considered to have unit
            strides.
        validate: Perform validation at the end of the function.
        validate_all: Perform validation also on intermediate steps.
    """
    return sdfg.apply_transformations_once_everywhere(
        MapIterationOrder(
            unit_strides_dims=unit_strides_dim,
            unit_strides_kind=unit_strides_kind,
        ),
        validate=validate,
        validate_all=validate_all,
    )


@dace_properties.make_properties
class MapIterationOrder(dace_transformation.SingleStateTransformation):
    """Modify the order of the iteration variables.

    The iteration order, while irrelevant from an SDFG point of view, is highly
    relevant in code and the fastest varying index ("inner most loop" in CPU or
    "x block dimension" in GPU) should be associated with the stride 1 dimension
    of the array.
    This transformation will reorder the map indexes such that this is the case.

    There are two ways to specify the unit stride dimensions. The first one is to
    specify `unit_strides_dim` at construction. It is possible to either pass a
    single dimension or a list of dimensions, in which case they are interpreted
    as priorities, the lower the index the higher its priority, see example below.
    Assuming we have the `unit_strides_dim=[EdgeDim, VertexDim]`, then we have the
    following:
    - `Map[EdgeDim, KDim, VertexDim]` -> `Map[KDim, VertexDim, EdgeDim]`.
    - `Map[VertexDim, KDim]` -> `Map[KDim, VertexDim]`.
    - `Map[EdgeDim, KDim]` -> `Map[KDim, EdgeDim]`.
    - `Map[CellDim, KDim]` -> `Map[CellDim, KDim]` (no modification).

    The second way is to specify `unit_strides_kind`, which is a GT4Py `DimensionKind`
    enum value. In that case the all dimensions that are associated to this kind
    will be ordered such that they are the inner most. If a Map has multiple
    dimensions of the same kind, then their relative order is undefined.

    In any case the order of the non matches dimensions is unspecific.

    Args:
        unit_strides_dims: GT4Py dimensions that are associated with the dimension
            that is supposed to have stride 1. If it is a list it is used as a ranking.
        unit_strides_kind: All dimensions of this kind have potential stride 1.

    Note:
        - The transformation does follow the rules outlined in
            [ADR0018](https://github.com/GridTools/gt4py/tree/main/docs/development/ADRs/0018-Canonical_SDFG_in_GT4Py_Transformations.md)
            especially rule 11, regarding the names. This is used to identify vertical
            and horizontal dimensions if `unit_strides_kind` is used.
        - String passed to `unit_strides_dim` have to match the name of the Map parameter
            exactly.

    Todo:
        - Maybe also process the parameters to bring them in a canonical order.
    """

    unit_strides_dims = dace_properties.ListProperty(
        element_type=str,
        allow_none=True,
        default=None,
        desc="Dimensions that should become the leading dimension.",
    )
    unit_strides_kind = dace_properties.Property(
        dtype=gtx_common.DimensionKind,
        allow_none=True,
        default=None,
        desc="Which kind of dimension should be made leading.",
    )
    map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    def __init__(
        self,
        unit_strides_dims: Optional[
            Union[str, gtx_common.Dimension, list[Union[str, gtx_common.Dimension]]]
        ] = None,
        unit_strides_kind: Optional[gtx_common.DimensionKind] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if unit_strides_dims is not None and unit_strides_kind is not None:
            raise ValueError("Specified both 'unit_strides_dims' and 'unit_strides_kind'.")
        elif unit_strides_dims is not None:
            if isinstance(unit_strides_dims, (gtx_common.Dimension, str)):
                unit_strides_dims = [unit_strides_dims]
            self.unit_strides_dims = [
                unit_strides_dim
                if isinstance(unit_strides_dim, str)
                else gtir_to_sdfg_utils.get_map_variable(unit_strides_dim)
                for unit_strides_dim in unit_strides_dims
            ]
        elif unit_strides_kind is not None:
            self.unit_strides_kind = unit_strides_kind
        else:
            raise ValueError("Neither specified 'unit_strides_kind' not 'unit_strides_dim'")

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.map_entry)]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        """Test if the map can be reordered.

        Essentially the function checks if the selected dimension is inside the map,
        and if so, if it is on the right place.
        """
        if self.compute_map_param_order() is None:
            return False
        return True

    def apply(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> None:
        """Performs the actual parameter reordering.

        The function will make the map variable, that corresponds to
        `self.unit_strides_dim` the last map variable (this is given by the structure of
        DaCe's code generator).
        """
        map_object: dace_nodes.Map = self.map_entry.map
        new_map_params_order: list[int] = self.compute_map_param_order()  # type: ignore[assignment]  # Guaranteed to be not `None`.

        def reorder(what: list[Any]) -> list[Any]:
            assert isinstance(what, list)
            return [what[new_pos] for new_pos in new_map_params_order]

        map_object.params = reorder(map_object.params)
        map_object.range.ranges = reorder(map_object.range.ranges)
        map_object.range.tile_sizes = reorder(map_object.range.tile_sizes)

    def compute_map_param_order(self) -> Optional[list[int]]:
        """Computes the new iteration order of the matched map.

        The function returns a list, the value at index `i` indicates the old dimension
        that should be put at the new location. If the order is already correct then
        `None` is returned.
        """
        map_entry: dace_nodes.MapEntry = self.map_entry
        map_params: list[str] = map_entry.map.params

        new_map_params = (
            self._compute_map_param_order_unit_strides_dims()
            if self.unit_strides_dims is not None
            else self._compute_map_param_order_unit_strides_kind()
        )

        if new_map_params is None or map_params == new_map_params:
            return None

        initial_order: dict[str, int] = {map_param: i for i, map_param in enumerate(map_params)}
        return [initial_order[new_map_param] for new_map_param in new_map_params]

    def _compute_map_param_order_unit_strides_kind(self) -> Optional[list[str]]:
        """Compute the new Map order for the case a `unit_strides_kind` was specified."""
        map_entry: dace_nodes.MapEntry = self.map_entry
        map_params: list[str] = map_entry.map.params
        kind_to_look_for: str = "_" + str(self.unit_strides_kind)

        # Because of the way how GT4Py generates the names, we can just check if the
        #  parameter ends with kind keyword. We will now divide the parameters into
        #  two groups the one we care and the one that we ignore.
        map_params_to_order: list[str] = [
            map_param for map_param in map_params if map_param.endswith(kind_to_look_for)
        ]
        if not map_params_to_order:
            return None

        # The final parameter order is that the "leading ones" are right most.
        #  However, their order, if there are multiple of the same kind, is unspecific.
        new_map_params: list[str] = [
            map_param for map_param in map_params if map_param not in map_params_to_order
        ]
        new_map_params.extend(map_params_to_order)

        return new_map_params

    def _compute_map_param_order_unit_strides_dims(self) -> Optional[list[str]]:
        """Computes the new iteration order if concrete dimensions were specified."""
        map_entry: dace_nodes.MapEntry = self.map_entry
        map_params: list[str] = map_entry.map.params

        # Find all the parameters that we want to order.
        map_params_to_order: set[str] = {
            map_param for map_param in map_params if map_param in self.unit_strides_dims
        }
        if not map_params_to_order:
            return None

        # We start with all parameters that we ignore/do not care about.
        new_map_params: list[str] = [
            map_param for map_param in map_params if map_param not in self.unit_strides_dims
        ]

        # Because how code generation works the leading dimension must be the most
        #  right one. Because this is also `self.unit_strides_dims[0]` we have to process
        #  then in reverse order.
        for map_param_to_check in reversed(self.unit_strides_dims):
            if map_param_to_check in map_params_to_order:
                new_map_params.append(map_param_to_check)
        return new_map_params



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import sys
from collections.abc import Callable
from typing import Any, Final, Iterable, Literal, Optional

from gt4py import eve
from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import builtins, ir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im


class ValidateRecordedShiftsAnnex(eve.NodeVisitor):
    """Ensure every applied lift and its arguments have the `recorded_shifts` annex populated."""

    def visit_FunCall(self, node: ir.FunCall):
        if cpm.is_applied_lift(node):
            assert hasattr(node.annex, "recorded_shifts")

            if len(node.annex.recorded_shifts) == 0:
                return

            if isinstance(node.fun.args[0], ir.Lambda):
                stencil = node.fun.args[0]
                for param in stencil.params:
                    assert hasattr(param.annex, "recorded_shifts")

        self.generic_visit(node)


def copy_recorded_shifts(from_: ir.Node, to: ir.Node) -> None:
    """
    Copy `recorded_shifts` annex attribute from one node to another.

    This function mainly exists for readability reasons.
    """
    to.annex.recorded_shifts = from_.annex.recorded_shifts


class Sentinel(eve.StrEnum):
    VALUE = "VALUE"
    TYPE = "TYPE"

    ALL_NEIGHBORS = "ALL_NEIGHBORS"


@dataclasses.dataclass(frozen=True)
class ShiftRecorder:
    recorded_shifts: dict[int, set[tuple[ir.OffsetLiteral, ...]]] = dataclasses.field(
        default_factory=dict
    )

    def register_node(self, inp: ir.Expr | ir.Sym) -> None:
        self.recorded_shifts.setdefault(id(inp), set())

    def __call__(self, inp: ir.Expr | ir.Sym, offsets: tuple[ir.OffsetLiteral, ...]) -> None:
        self.recorded_shifts[id(inp)].add(offsets)


@dataclasses.dataclass(frozen=True)
class ForwardingShiftRecorder:
    wrapped_tracer: Any
    shift_recorder: ShiftRecorder

    def __call__(self, inp: ir.Expr | ir.Sym, offsets: tuple[ir.OffsetLiteral, ...]):
        self.shift_recorder(inp, offsets)
        # Forward shift to wrapped tracer such it can record the shifts of the parent nodes
        self.wrapped_tracer.shift(offsets).deref()


# for performance reasons (`isinstance` is slow otherwise) we don't use abc here
class Tracer:
    def deref(self):
        raise NotImplementedError()

    def shift(self, offsets: tuple[ir.OffsetLiteral, ...]):
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class ArgTracer(Tracer):
    arg: ir.Expr | ir.Sym
    shift_recorder: ShiftRecorder | ForwardingShiftRecorder
    offsets: tuple[ir.OffsetLiteral, ...] = ()

    def shift(self, offsets: tuple[ir.OffsetLiteral, ...]):
        return ArgTracer(
            arg=self.arg, shift_recorder=self.shift_recorder, offsets=self.offsets + tuple(offsets)
        )

    def deref(self):
        self.shift_recorder(self.arg, self.offsets)
        return Sentinel.VALUE


# This class is only needed because we currently allow conditionals on iterators. Since this is
# not supported in the C++ backend it can likely be removed again in the future.
@dataclasses.dataclass(frozen=True)
class CombinedTracer(Tracer):
    its: tuple[Tracer, ...]

    def shift(self, offsets: tuple[ir.OffsetLiteral, ...]):
        return CombinedTracer(tuple(_shift(*offsets)(it) for it in self.its))

    def deref(self):
        derefed_its = [it.deref() for it in self.its]
        if not all(it == Sentinel.VALUE for it in derefed_its[1:]):
            raise AssertionError("The result of a `deref` must be a `Sentinel.VALUE`.")
        return Sentinel.VALUE


def _combine(*values):
    # `OffsetLiteral`s may occur in `list_get` calls
    if not all(
        val in [Sentinel.VALUE, Sentinel.TYPE] or isinstance(val, ir.OffsetLiteral)
        for val in values
    ):
        raise AssertionError("All arguments must be values or types.")
    return Sentinel.VALUE


# implementations of builtins
def _deref(x):
    return x.deref()


def _can_deref(x):
    return Sentinel.VALUE


def _shift(*offsets):
    assert all(
        isinstance(offset, ir.OffsetLiteral) or offset in [Sentinel.ALL_NEIGHBORS, Sentinel.VALUE]
        for offset in offsets
    )

    def apply(arg):
        assert isinstance(arg, Tracer)
        return arg.shift(offsets)

    return apply


@dataclasses.dataclass(frozen=True)
class AppliedLift(Tracer):
    stencil: Callable
    its: tuple[Tracer, ...]

    def shift(self, offsets):
        return AppliedLift(self.stencil, tuple(_shift(*offsets)(it) for it in self.its))

    def deref(self):
        return self.stencil(*self.its)


def _lift(f):
    def apply(*its):
        if not all(isinstance(it, Tracer) for it in its):
            raise AssertionError("All arguments must be iterators.")
        return AppliedLift(f, its)

    return apply


def _reduce(f, init):
    return _combine


def _map(f):
    return _combine


def _neighbors(o, x):
    return _deref(_shift(o, Sentinel.ALL_NEIGHBORS)(x))


def _scan(f, forward, init):
    def apply(*args):
        return f(init, *args)

    return apply


def _primitive_constituents(
    val: Literal[Sentinel.VALUE] | Tracer | tuple,
) -> Iterable[Literal[Sentinel.VALUE] | Tracer]:
    if val is Sentinel.VALUE or isinstance(val, Tracer):
        yield val
    elif isinstance(val, tuple):
        for el in val:
            if isinstance(el, tuple):
                yield from _primitive_constituents(el)
            elif el is Sentinel.VALUE or isinstance(el, Tracer):
                yield el
            else:
                raise AssertionError("Expected a `Sentinel.VALUE`, `Tracer` or tuple thereof.")
    else:
        raise ValueError()


def _if(cond: Literal[Sentinel.VALUE], true_branch, false_branch):
    assert cond is Sentinel.VALUE
    if any(isinstance(branch, tuple) for branch in (false_branch, true_branch)):
        # Broadcast branches to tuple of same length. This is required for cases like:
        #  `if_(cond, deref(iterator_of_tuples), make_tuple(...))`.
        if not isinstance(true_branch, tuple):
            assert all(el == Sentinel.VALUE for el in false_branch)
            true_branch = (true_branch,) * len(false_branch)
        if not isinstance(false_branch, tuple):
            assert all(el == Sentinel.VALUE for el in true_branch)
            false_branch = (false_branch,) * len(true_branch)

        result = []
        for el_true_branch, el_false_branch in zip(true_branch, false_branch):
            # just reuse `if_` to recursively build up the result
            result.append(_if(Sentinel.VALUE, el_true_branch, el_false_branch))
        return tuple(result)

    is_iterator_arg = tuple(isinstance(arg, Tracer) for arg in (cond, true_branch, false_branch))
    if is_iterator_arg == (False, True, True):
        return CombinedTracer((true_branch, false_branch))
    assert is_iterator_arg == (False, False, False) and all(
        arg in [Sentinel.VALUE, Sentinel.TYPE] for arg in (cond, true_branch, false_branch)
    )
    return Sentinel.VALUE


def _make_tuple(*args):
    return args


def _tuple_get(index, tuple_val):
    if isinstance(tuple_val, tuple):
        return tuple_val[index]
    assert tuple_val is Sentinel.VALUE
    return Sentinel.VALUE


def _as_fieldop(stencil, domain=None):
    def applied_as_fieldop(*args):
        return stencil(*args)

    return applied_as_fieldop


_START_CTX: Final = {
    "as_fieldop": _as_fieldop,
    "deref": _deref,
    "can_deref": _can_deref,
    "shift": _shift,
    "lift": _lift,
    "scan": _scan,
    "reduce": _reduce,
    "neighbors": _neighbors,
    "map_": _map,
    "if_": _if,
    "make_tuple": _make_tuple,
}


# TODO(tehrengruber): This pass is unnecessarily very inefficient and easily exceeds the default
#  recursion limit.
@dataclasses.dataclass(frozen=True)
class TraceShifts(PreserveLocationVisitor, NodeTranslator):
    shift_recorder: ShiftRecorder = dataclasses.field(default_factory=ShiftRecorder)

    def visit_Literal(self, node: ir.SymRef, *, ctx: dict[str, Any]) -> Any:
        return Sentinel.VALUE

    def visit_InfinityLiteral(self, node: ir.InfinityLiteral, *, ctx: dict[str, Any]):
        return Sentinel.VALUE

    def visit_SymRef(self, node: ir.SymRef, *, ctx: dict[str, Any]) -> Any:
        if node.id in ctx:
            return ctx[node.id]
        elif node.id in builtins.TYPE_BUILTINS:
            return Sentinel.TYPE
        elif node.id in (builtins.ARITHMETIC_BUILTINS | {"list_get", "make_const_list", "cast_"}):
            return _combine
        raise ValueError(f"Undefined symbol {node.id}")

    def visit_FunCall(self, node: ir.FunCall, *, ctx: dict[str, Any]) -> Any:
        if node.fun == ir.SymRef(id="tuple_get"):
            assert isinstance(node.args[0], ir.Literal)
            index = int(node.args[0].value)
            return _tuple_get(index, self.visit(node.args[1], ctx=ctx))

        fun = self.visit(node.fun, ctx=ctx)
        args = self.visit(node.args, ctx=ctx)
        return fun(*args)

    def visit(self, node, **kwargs):
        result = super().visit(node, **kwargs)
        if isinstance(result, Tracer):
            assert isinstance(node, (ir.Sym, ir.Expr))

            self.shift_recorder.register_node(node)
            result = ArgTracer(
                arg=node, shift_recorder=ForwardingShiftRecorder(result, self.shift_recorder)
            )
        return result

    def visit_Lambda(self, node: ir.Lambda, *, ctx: dict[str, Any]) -> Callable:
        def fun(*args):
            new_args = []
            for param, arg in zip(node.params, args, strict=True):
                if isinstance(arg, Tracer):
                    self.shift_recorder.register_node(param)
                    new_args.append(
                        ArgTracer(
                            arg=param,
                            shift_recorder=ForwardingShiftRecorder(arg, self.shift_recorder),
                        )
                    )
                else:
                    new_args.append(arg)

            return self.visit(
                node.expr, ctx=ctx | {p.id: a for p, a in zip(node.params, new_args, strict=True)}
            )

        return fun

    @classmethod
    def trace_stencil(
        cls, stencil: ir.Expr, *, num_args: Optional[int] = None, save_to_annex: bool = False
    ) -> list[set[tuple[ir.OffsetLiteral, ...]]]:
        # If we get a lambda we can deduce the number of arguments.
        if isinstance(stencil, ir.Lambda):
            assert num_args is None or num_args == len(stencil.params)
            num_args = len(stencil.params)
        elif cpm.is_call_to(stencil, "scan"):
            assert isinstance(stencil.args[0], ir.Lambda)
            num_args = len(stencil.args[0].params) - 1
        if not isinstance(num_args, int):
            raise ValueError("Stencil must be an 'itir.Lambda', scan, or `num_args` is given.")
        assert isinstance(num_args, int)

        args = [im.ref(f"__arg{i}") for i in range(num_args)]

        old_recursionlimit = sys.getrecursionlimit()
        sys.setrecursionlimit(100000000)

        instance = cls()

        # initialize shift recorder & context with all built-ins and the iterator argument tracers
        ctx: dict[str, Any] = {**_START_CTX}
        for arg in args:
            instance.shift_recorder.register_node(arg)
            ctx[arg.id] = ArgTracer(arg=arg, shift_recorder=instance.shift_recorder)

        # actually trace stencil
        instance.visit(im.call(stencil)(*args), ctx=ctx)

        sys.setrecursionlimit(old_recursionlimit)

        recorded_shifts = instance.shift_recorder.recorded_shifts

        param_shifts = []
        for arg in args:
            param_shifts.append(recorded_shifts[id(arg)])

        if save_to_annex:
            _save_to_annex(stencil, recorded_shifts)

        return param_shifts


trace_stencil = TraceShifts.trace_stencil


def _save_to_annex(
    node: ir.Node, recorded_shifts: dict[int, set[tuple[ir.OffsetLiteral, ...]]]
) -> None:
    for child_node in node.pre_walk_values():
        if id(child_node) in recorded_shifts:
            child_node.annex.recorded_shifts = recorded_shifts[id(child_node)]

    if __debug__:
        ValidateRecordedShiftsAnnex().visit(node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Contains the exception classes and other utilities for error handling."""

from . import excepthook  # noqa: F401 [unused-import]
from .exceptions import (
    CompilationError,
    DSLError,
    DSLTypeError,
    InvalidParameterAnnotationError,
    MissingArgumentError,
    MissingAttributeError,
    MissingParameterAnnotationError,
    UndefinedSymbolError,
    UnsupportedPythonFeatureError,
)


__all__ = [
    "CompilationError",
    "DSLError",
    "DSLTypeError",
    "InvalidParameterAnnotationError",
    "MissingArgumentError",
    "MissingAttributeError",
    "MissingParameterAnnotationError",
    "UndefinedSymbolError",
    "UnsupportedPythonFeatureError",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


import numbers

import numpy as np


class Field:
    def __init__(
        self, field: np.ndarray, offsets: tuple[int, ...], dimensions: tuple[bool, bool, bool]
    ):
        ii = iter(range(3))
        self.idx_to_data = tuple(
            [next(ii) if has_dim else None for has_dim in dimensions]
            + list(range(sum(dimensions), len(field.shape)))
        )

        shape = [field.shape[i] if i is not None else 1 for i in self.idx_to_data]
        self.field_view = np.reshape(field.data, shape).view(np.ndarray)

        self.offsets = offsets

    @classmethod
    def empty(cls, shape, dtype, offset):
        return cls(np.empty(shape, dtype=dtype), offset, (True, True, True))

    def shim_key(self, key):
        new_args = []
        if not isinstance(key, tuple):
            key = (key,)
        for index in self.idx_to_data:
            if index is None:
                new_args.append(slice(None, None))
            else:
                idx = key[index]
                offset = self.offsets[index]
                if isinstance(idx, slice):
                    new_args.append(
                        slice(idx.start + offset, idx.stop + offset, idx.step) if offset else idx
                    )
                else:
                    new_args.append(idx + offset)
        if not isinstance(new_args[2], (numbers.Integral, slice)):
            new_args = self.broadcast_and_clip_variable_k(new_args)
        return tuple(new_args)

    def broadcast_and_clip_variable_k(self, new_args: list):
        assert isinstance(new_args[0], slice) and isinstance(new_args[1], slice)
        if np.max(new_args[2]) >= self.field_view.shape[2] or np.min(new_args[2]) < 0:
            new_args[2] = np.clip(new_args[2].copy(), 0, self.field_view.shape[2] - 1)
        new_args[:2] = np.broadcast_arrays(
            np.expand_dims(
                np.arange(new_args[0].start, new_args[0].stop),
                axis=tuple(i for i in range(self.field_view.ndim) if i != 0),
            ),
            np.expand_dims(
                np.arange(new_args[1].start, new_args[1].stop),
                axis=tuple(i for i in range(self.field_view.ndim) if i != 1),
            ),
        )
        return new_args

    def __getitem__(self, key):
        return self.field_view.__getitem__(self.shim_key(key))

    def __setitem__(self, key, value):
        return self.field_view.__setitem__(self.shim_key(key), value)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import multiprocessing
import os
from typing import Any, Dict, List, Optional

import gridtools_cpp

from gt4py._core import definitions as core_defs


GT4PY_INSTALLATION_PATH: str = os.path.dirname(os.path.abspath(__file__))

CUDA_ROOT: str = os.environ.get(
    "CUDA_HOME", os.environ.get("CUDA_PATH", os.path.abspath("/usr/local/cuda"))
)

CUDA_HOST_CXX: Optional[str] = os.environ.get("CUDA_HOST_CXX", None)

GT_INCLUDE_PATH: str = os.path.abspath(gridtools_cpp.get_include_dir())

GT_CPP_TEMPLATE_DEPTH: int = 1024

GT4PY_COMPILE_OPT_LEVEL: str = os.environ.get("GT4PY_COMPILE_OPT_LEVEL", "3")
GT4PY_EXTRA_COMPILE_OPT_FLAGS: str = os.environ.get("GT4PY_EXTRA_COMPILE_OPT_FLAGS", "")

# Settings dict
GT4PY_EXTRA_COMPILE_ARGS: str = os.environ.get("GT4PY_EXTRA_COMPILE_ARGS", "")
extra_compile_args: List[str] = (
    list(GT4PY_EXTRA_COMPILE_ARGS.split(" ")) if GT4PY_EXTRA_COMPILE_ARGS else []
)
GT4PY_EXTRA_LINK_ARGS: str = os.environ.get("GT4PY_EXTRA_LINK_ARGS", "")
extra_link_args: List[str] = list(GT4PY_EXTRA_LINK_ARGS.split(" ")) if GT4PY_EXTRA_LINK_ARGS else []

build_settings: Dict[str, Any] = {
    "cuda_bin_path": os.path.join(CUDA_ROOT, "bin"),
    "cuda_include_path": os.path.join(CUDA_ROOT, "include"),
    "cuda_arch": os.environ.get("CUDA_ARCH", None),
    "gt_include_path": os.environ.get("GT_INCLUDE_PATH", GT_INCLUDE_PATH),
    "openmp_cppflags": os.environ.get("OPENMP_CPPFLAGS", "-fopenmp").split(),
    "openmp_ldflags": os.environ.get("OPENMP_LDFLAGS", "-fopenmp").split(),
    "extra_compile_args": {"cxx": extra_compile_args, "cuda": extra_compile_args},
    "extra_link_args": extra_link_args,
    "parallel_jobs": multiprocessing.cpu_count(),
    "cpp_template_depth": os.environ.get("GT_CPP_TEMPLATE_DEPTH", GT_CPP_TEMPLATE_DEPTH),
}
if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
    build_settings["cuda_library_path"] = os.path.join(CUDA_ROOT, "lib")
else:
    build_settings["cuda_library_path"] = os.path.join(CUDA_ROOT, "lib64")

if CUDA_HOST_CXX is not None:
    build_settings["extra_compile_args"]["cuda"].append(f"-ccbin={CUDA_HOST_CXX}")

cache_settings: Dict[str, Any] = {
    "dir_name": os.environ.get("GT_CACHE_DIR_NAME", ".gt_cache"),
    "root_path": os.environ.get("GT_CACHE_ROOT", os.path.abspath(".")),
    "load_retries": int(os.environ.get("GT_CACHE_LOAD_RETRIES", 3)),
    "load_retry_delay": int(os.environ.get("GT_CACHE_LOAD_RETRY_DELAY", 100)),  # unit milliseconds
}

code_settings: Dict[str, Any] = {"root_package_name": "_GT_"}

os.environ.setdefault("DACE_CONFIG", os.path.join(os.path.abspath("."), ".dace.conf"))

DACE_DEFAULT_BLOCK_SIZE: str = os.environ.get("DACE_DEFAULT_BLOCK_SIZE", "64,8,1")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

__all__ = ["StencilTestSuite", "field", "global_name", "none", "parameter"]
try:
    from .input_strategies import field, global_name, none, parameter
    from .suites import StencilTestSuite
except ModuleNotFoundError as e:
    if "pytest" in str(e) or "hypothesis" in str(e):
        raise RuntimeError(
            "Missing 'testing' extra dependencies. Reinstall GT4Py as 'gt4py[testing]'"
        ) from e
    else:
        raise e



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from typing import Any, Generic, List, TypeAlias, TypeGuard, TypeVar

from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import ir_makers as im


_Fun = TypeVar("_Fun", bound=itir.Expr)


class _FunCallTo(itir.FunCall, Generic[_Fun]):
    fun: _Fun
    args: List[itir.Expr]


_FunCallToSymRef: TypeAlias = _FunCallTo[itir.SymRef]


def is_call_to(node: Any, fun: str | Iterable[str]) -> TypeGuard[_FunCallToSymRef]:
    """
    Match call expression to a given function.

    If the `node` argument is not an `itir.Node` the function does not error, but just returns
    `False`. This is useful in visitors, where sometimes we pass a list of nodes or a leaf
    attribute which can be anything.

    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> node = im.plus(1, 2)
    >>> is_call_to(node, "plus")
    True
    >>> is_call_to(node, "minus")
    False
    >>> is_call_to(node, ("plus", "minus"))
    True
    """
    assert not isinstance(fun, itir.Node)  # to avoid accidentally passing the fun as first argument
    if isinstance(fun, str):
        return (
            isinstance(node, itir.FunCall)
            and isinstance(node.fun, itir.SymRef)
            and node.fun.id == fun
        )
    else:
        return any((is_call_to(node, f) for f in fun))


_FunCallToFunCallToRef: TypeAlias = _FunCallTo[_FunCallToSymRef]


def is_applied_lift(arg: itir.Node) -> TypeGuard[_FunCallToFunCallToRef]:
    """Match expressions of the form `lift(λ(...) → ...)(...)`."""
    return (
        isinstance(arg, itir.FunCall)
        and isinstance(arg.fun, itir.FunCall)
        and isinstance(arg.fun.fun, itir.SymRef)
        and arg.fun.fun.id == "lift"
    )


def is_applied_map(arg: itir.Node) -> TypeGuard[_FunCallToFunCallToRef]:
    """Match expressions of the form `map(λ(...) → ...)(...)`."""
    return (
        isinstance(arg, itir.FunCall)
        and isinstance(arg.fun, itir.FunCall)
        and isinstance(arg.fun.fun, itir.SymRef)
        and arg.fun.fun.id == "map_"
    )


def is_applied_reduce(arg: itir.Node) -> TypeGuard[_FunCallToFunCallToRef]:
    """Match expressions of the form `reduce(λ(...) → ...)(...)`."""
    return (
        isinstance(arg, itir.FunCall)
        and isinstance(arg.fun, itir.FunCall)
        and isinstance(arg.fun.fun, itir.SymRef)
        and arg.fun.fun.id == "reduce"
    )


def is_applied_shift(arg: itir.Node) -> TypeGuard[_FunCallToFunCallToRef]:
    """Match expressions of the form `shift(λ(...) → ...)(...)`."""
    return (
        isinstance(arg, itir.FunCall)
        and isinstance(arg.fun, itir.FunCall)
        and isinstance(arg.fun.fun, itir.SymRef)
        and arg.fun.fun.id == "shift"
    )


def is_applied_as_fieldop(arg: itir.Node) -> TypeGuard[_FunCallToFunCallToRef]:
    """Match expressions of the form `as_fieldop(stencil)(*args)`."""
    return isinstance(arg, itir.FunCall) and is_call_to(arg.fun, "as_fieldop")


_FunCallToLambda: TypeAlias = _FunCallTo[itir.Lambda]


def is_let(node: itir.Node) -> TypeGuard[_FunCallToLambda]:
    """Match expression of the form `(λ(...) → ...)(...)`."""
    return isinstance(node, itir.FunCall) and isinstance(node.fun, itir.Lambda)


def is_ref_to(node, ref: str) -> TypeGuard[itir.SymRef]:
    return isinstance(node, itir.SymRef) and node.id == ref


def is_identity_as_fieldop(node: itir.Expr) -> TypeGuard[_FunCallToFunCallToRef]:
    """
    Match field operators implementing element-wise copy of a field argument,
    that is expressions of the form `as_fieldop(stencil)(*args)`

    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> node = im.as_fieldop(im.lambda_("__arg0")(im.deref("__arg0")))("a")
    >>> is_identity_as_fieldop(node)
    True
    >>> node = im.as_fieldop("deref")("a")
    >>> is_identity_as_fieldop(node)
    False
    """
    if not is_applied_as_fieldop(node):
        return False
    stencil = node.fun.args[0]
    if (
        isinstance(stencil, itir.Lambda)
        and len(stencil.params) == 1
        and stencil == im.lambda_(stencil.params[0])(im.deref(stencil.params[0].id))
    ):
        return True
    return False



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from typing import Any

import dace
from dace import transformation as dace_transformation
from dace.sdfg import nodes as dace_nodes, utils as dace_sdutils


# Conditional import because `gt4py.cartesian` uses an older DaCe version without
#  `explicit_cf_compatible`.
# TODO(phimuell): Remove once `gt4py.cartesian` has been updated.
try:
    explicit_cf_compatible = dace_transformation.explicit_cf_compatible
except AttributeError:
    explicit_cf_compatible = lambda x: x  # noqa: E731 [lambda-assignment]


@explicit_cf_compatible
class GT4PyStateFusion(dace_transformation.MultiStateTransformation):
    """Implements a state fusion transformation that is specific to GT4Py.

    Because there are, by construction, no write conflicts in transient this
    transformation can be much simpler, than the one in DaCe.
    It should not be seen as a replacement but as an extension.

    The transformation can only be applied in the following cases:
    - The first state has only one outgoing edge, that connects it with the second state.
    - The second only has one incoming edge, that connects it with the first state.
    - The connecting edge is unconditional and does not contain any assignments.
    - If global data is written to in the first state and there is an AccessNode that
        reads and writes to the same data in the second state the transformation
        does not apply.
    - No read writes conflicts are created.

    If the conditions above are met, then the transformation will copy the nodes from
    the second state into the first state and if needed create connections as needed.

    Todo:
        - Add a step that merges global AccessNodes together, this is borderline
            ADR-18 compatibility.
        - If both states read from the same transient data, that was not defined in
            either, then the AccessNodes are not merged.
        - Refactor this transformation such that it pre compute more.
    """

    first_state = dace_transformation.PatternNode(dace.SDFGState)
    second_state = dace_transformation.PatternNode(dace.SDFGState)

    @classmethod
    def expressions(cls) -> Any:
        return [dace_sdutils.node_path_graph(cls.first_state, cls.second_state)]

    @staticmethod
    def annotates_memlets() -> bool:
        return False

    def can_be_applied(
        self, _: Any, expr_index: int, sdfg: dace.SDFG, permissive: bool = False
    ) -> bool:
        first_state: dace.SDFGState = self.first_state
        second_state: dace.SDFGState = self.second_state
        graph: dace.sdfg.state.AbstractControlFlowRegion = first_state.parent_graph

        if graph.out_degree(first_state) != 1:
            return False
        if graph.in_degree(second_state) != 1:
            return False
        conn_edge = next(iter(graph.out_edges(first_state)))
        if not conn_edge.data.is_unconditional():
            return False

        # TODO(phimuell): Lift this limitation.
        if len(conn_edge.data.assignments) != 0:
            return False

        # If one of the states is empty, then we can accept it because it will not
        #  create any data access issues.
        if first_state.number_of_nodes() == 0:
            return True
        if second_state.number_of_nodes() == 0:
            return True

        # If the first state writes to global memory and the second state contains
        #  an AccessNode that reads and writes to the same global memory. In this case
        #  we do not know how to merge the nodes together. Thus we reject these cases
        # NOTE: This is a very obscure case.
        first_global_memory_write: set[str] = {
            dnode.data
            for dnode in first_state.data_nodes()
            if not dnode.desc(sdfg).transient and first_state.in_degree(dnode) != 0
        }
        if any(
            dnode.data in first_global_memory_write
            for dnode in second_state.data_nodes()
            if second_state.in_degree(dnode) != 0 and second_state.out_degree(dnode) != 0
        ):
            return False

        if self._check_for_read_write_conflicts(sdfg):
            return False

        return True

    def _check_for_read_write_conflicts(
        self,
        sdfg: dace.SDFG,
    ) -> bool:
        """Return `True` if there are conflicts."""

        if self._check_for_read_write_dependencies(sdfg):
            return True
        if self._check_for_wcr_conflicts(sdfg):
            return True
        return False

    def _check_for_wcr_conflicts(
        self,
        sdfg: dace.SDFG,
    ) -> bool:
        """Checks if wcr edges prevent two consecutive states to be fused.

        If both states contains a write to the same data, and one of them has a set
        `wcr` then we can not merge the states.

        Note:
            In case all writes are `wcr` it might be possible to merge them, but
            this is a rather obscure case that we ignore.
        """
        first_state: dace.SDFGState = self.first_state
        first_scope_dict = first_state.scope_dict()
        second_state: dace.SDFGState = self.second_state
        second_scope_dict = second_state.scope_dict()

        # Collect the AccessNodes that are written to in the first state.
        first_state_writes: list[dace_nodes.AccessNode] = [
            dnode
            for dnode in first_state.data_nodes()
            if first_scope_dict[dnode] is None and first_state.in_degree(dnode) != 0
        ]
        first_state_writes_names: set[str] = {ac.data for ac in first_state_writes}

        # Now collect all the AccessNodes that are written to in the second state.
        second_state_writes: list[dace_nodes.AccessNode] = [
            dnode
            for dnode in second_state.data_nodes()
            if second_scope_dict[dnode] is None and second_state.in_degree(dnode) != 0
        ]
        second_state_writes_names: set[str] = {ac.data for ac in second_state_writes}

        # If there is no data that is written to in both state then there is no conflict.
        common_write_data: set[str] = first_state_writes_names.intersection(
            second_state_writes_names
        )
        if len(common_write_data) == 0:
            return False

        for state, ac_nodes in [
            (first_state, first_state_writes),
            (second_state, second_state_writes),
        ]:
            for ac in ac_nodes:
                if ac.data not in common_write_data:
                    continue
                if any(iedge.data.wcr is not None for iedge in state.in_edges(ac)):
                    return True

        return False

    def _check_for_read_write_dependencies(
        self,
        sdfg: dace.SDFG,
    ) -> bool:
        """Checks for read write conflicts in the global memory.

        The function checks if by the elimination of states creates read write
        conflicts. Because of the structure outlined by ADR-18 it only checks this
        for the global data, since transients are by definition written only once.

        Todo:
            Refine this function.
        """
        first_state: dace.SDFGState = self.first_state
        first_scope_dict = first_state.scope_dict()
        second_state: dace.SDFGState = self.second_state
        second_scope_dict = second_state.scope_dict()

        # Determine the concurrent subgraphs, i.e. the data flow that can, in theory
        #  execute in parallel. Note that this is not the same as connected components.
        #  As subgraphs might share nodes, although this should happen only for source
        #  nodes. For each of them we determine the set of data they write and on
        #  which data they depend on.
        first_subgraphs = dace.sdfg.utils.concurrent_subgraphs(first_state)
        data_producers: list[set[str]] = []
        data_producers_dependencies: list[set[str]] = []
        all_data_producers: set[str] = set()
        for first_subgraph in first_subgraphs:
            data_producers.append(
                {
                    dnode.data
                    for dnode in first_subgraph.data_nodes()
                    if first_scope_dict[dnode] is None and first_subgraph.in_degree(dnode) != 0
                }
            )
            data_producers_dependencies.append(
                {
                    dnode.data
                    for dnode in first_subgraph.data_nodes()
                    if first_scope_dict[dnode] is None and first_subgraph.out_degree(dnode) != 0
                }
            )
            if not all_data_producers.isdisjoint(data_producers[-1]):
                # TODO(phimuell): It is certainly possible to lift this restriction. A
                #   possible cause is that we found once a source node and then an
                #   AccessNode to the same data. In case it is global this is most likely
                #   valid. However, I think that simply allow it, is not okay, because
                #   it might break some assumption in the fuse code.
                warnings.warn(
                    f"Detected that '{first_state}' writes to the data"
                    f" `{', '.join(all_data_producers.intersection(data_producers[-1]))}`"
                    " in multiple concurrent subgraphs. This might indicate an error.",
                    stacklevel=0,
                )
                return True
            all_data_producers.update(data_producers[-1])

        # Now determine the concurrent subgraphs of the second state, i.e. the parts
        #  that in theory could execute in parallel. For every subgraph we determine
        #  which data it reads, that were written to in the first state, the so called
        #  messenger data. Components that share a messenger data will be merged
        #  together, if the states are fused. Furthermore, we determine the set of data
        #  a component writes to. Because if a component, from the first state, depends
        #  on data a component from the second state writes to and the two components
        #  are not related through a messenger data, then we end up with two
        #  independent subgraphs that read and write from the same global data.
        #  This is indeterministic behaviour, thus we should reject it.
        second_subgraphs = dace.sdfg.utils.concurrent_subgraphs(second_state)
        data_consumers: list[set[str]] = []
        data_consumers_influences: list[set[str]] = []
        messanger_data: set[str] = set()
        for second_subgraph in second_subgraphs:
            data_consumers.append(
                {
                    dnode.data
                    for dnode in second_subgraph.data_nodes()
                    if second_scope_dict[dnode] is None and dnode.data in all_data_producers
                }
            )
            data_consumers_influences.append(
                {
                    dnode.data
                    for dnode in second_subgraph.data_nodes()
                    if second_scope_dict[dnode] is None and second_subgraph.in_degree(dnode) != 0
                }
            )
            messanger_data.update(data_consumers[-1].intersection(all_data_producers))

        for data_producer, producer_dependency in zip(data_producers, data_producers_dependencies):
            produced_messenger = data_producer.intersection(messanger_data)

            if len(produced_messenger) == 0:
                # The component does not generate any messenger data. In that case
                #  we must ensure that it does not depend on any data that is
                #  modified by the second state.
                if not all(
                    producer_dependency.isdisjoint(consumer_influece)
                    for consumer_influece in data_consumers_influences
                ):
                    return True
                continue

            # The ID of the consumers that depend on the current producer. Note that
            #  we do not consider consumers here that fully depends on the producer.
            depending_consumers: list[int] = []

            for consumer_id, (data_consumer, consumer_influece) in enumerate(
                zip(data_consumers, data_consumers_influences)
            ):
                consumed_messenger = data_consumer.intersection(messanger_data)
                exchanged_messenger = produced_messenger.intersection(consumed_messenger)
                if len(exchanged_messenger) == 0:
                    # This component does not consume anything the first state
                    #  generates, this means it is an isolated component and remains
                    #  an isolated component. There is technically nothing to do.
                    #  However, there is the special case that it writes a variable
                    #  some component of the first state depends on. Strictly speaking
                    #  this violates ADR-18, but it happens in some scan expressions.
                    # We do not add the consumer to `depending_consumer` because there
                    #  is no connection.
                    if not producer_dependency.isdisjoint(consumer_influece):
                        # The consumer modifies something the producer needs. So we
                        #  can not fuse it.
                        return True

                else:
                    # There is some exchange between the producer and the consumer.
                    #  We assume that despite the merge the inner dependencies are
                    #  meet, i.e. all messenger materializes before the dependent
                    #  dataflow is executed. We thus have to ensure that the new
                    #  component will not affect any component that is not involved
                    #  inside the merge.
                    unaffected_producers: list[int] = [
                        producer_id
                        for producer_id, data_producer in enumerate(data_producers)
                        if data_producer.isdisjoint(data_consumer)
                    ]
                    dependencies_of_unaffected_producers: set[str] = set()
                    for unaffected_producer in unaffected_producers:
                        dependencies_of_unaffected_producers.update(
                            data_producers_dependencies[unaffected_producer]
                        )
                    if not dependencies_of_unaffected_producers.isdisjoint(consumer_influece):
                        return True

                    # In case the consumer did not consume everything inside the
                    #  producer add it to the list of dependent consumers. See
                    #  below for more.
                    # TODO(phimuell): Lift this or refine this.
                    if exchanged_messenger != produced_messenger:
                        depending_consumers.append(consumer_id)

            if len(depending_consumers) > 1:
                # If we have more than one depending consumer, then this means that
                #  several consumers would partially read the producer. This
                #  essentially creates concurrent dataflow within the component, which
                #  would lead to non-deterministic results, so we have to reject the
                #  merge.
                return True

        return False

    def _move_nodes(self, sdfg: dace.SDFG) -> None:
        """Moves the nodes from the second state to the first state.

        The function will assign all nodes from the second state into the first state.
        It will also modify the data flow to preserve the order of execution. This
        means if a data container was written to in the first state and read in the
        second state, then the reads to the data will not use the AccessNode that was
        in the second state, instead they will be rerouted such that they use the
        one from the first state.
        It is important that this function will not mutate the second state, thus
        this function will produce an SDFG in an invalid state.
        """
        first_state: dace.SDFGState = self.first_state
        second_state: dace.SDFGState = self.second_state
        assert isinstance(first_state, dace.SDFGState)
        assert isinstance(second_state, dace.SDFGState)

        first_scope_dict = first_state.scope_dict()
        second_scope_dict = second_state.scope_dict()

        # We have to preserve the order this means that every node from the second
        #  state, must merge with the corresponding node from the first state.
        #  These are all nodes that have a non zero input degree. For convenience,
        #  we also add all other data that is read, except if it is not written
        # TODO(phimuell): In case of global data it might be possible that there are
        #   multiple AccessNodes that writes to the data. We currently ignore that case.
        data_producers: dict[str, dace_nodes.AccessNode] = {
            dnode.data: dnode
            for dnode in first_state.data_nodes()
            if (first_scope_dict[dnode] is None and first_state.in_degree(dnode) != 0)
        }

        # Now add everything that is read too, if it is not already present.
        for dnode in first_state.data_nodes():
            if not (first_state.in_degree(dnode) == 0 and first_state.out_degree(dnode) != 0):
                continue
            if dnode.data in data_producers:
                assert not dnode.desc(sdfg).transient
                continue
            data_producers[dnode.data] = dnode

        # Now we will look for all data consumers, i.e. nodes that are reading data
        #  from the first state. These are all AccessNodes that have a zero indegree
        #  and that are listed in `data_producers`. Note that these nodes might have
        #  to be replaced with the nodes from the first state.
        #  Note we can not use a `dict` here because it is possible, although not
        #  fully compliant to ADR-18 though, that there are multiple AccessNodes
        #  referring to the same data.
        data_consumers: list[dace_nodes.AccessNode] = [
            dnode
            for dnode in second_state.data_nodes()
            if (
                second_scope_dict[dnode] is None
                and second_state.in_degree(dnode) == 0
                and dnode.data in data_producers
            )
        ]

        # Move the nodes and edges into from the second into the first state. However
        #  they are still part of the second state, this will be handled afterwards.
        for node in second_state.nodes():
            if isinstance(node, dace_nodes.NestedSDFG):
                node.sdfg.parent = first_state
            first_state.add_node(node)
        for src, src_conn, dst, dst_conn, data in second_state.edges():
            first_state.add_edge(src, src_conn, dst, dst_conn, data)

        # The first state generated data that the second state consumed, since they
        #  are now in one single state, we have to ensure an order. For this we simply
        #  merge the two nodes, which means that instead of reading from the original
        #  AccessNode (that was part of the second state) it should now read from the
        #  AccessNode that was defined in the first state. If there were multiple
        #  AccessNodes in the second state, that refer to the same data, they are
        #  all merged.
        for data_consumer in data_consumers:
            data_producer: dace_nodes.AccessNode = data_producers[data_consumer.data]

            # Now modify the edge, there is no other modification needed.
            for old_edge in first_state.out_edges(data_consumer):
                first_state.add_edge(
                    data_producer, old_edge.src_conn, old_edge.dst, old_edge.dst_conn, old_edge.data
                )
            first_state.remove_node(data_consumer)

        # For ADR-18 compatibility we now have to ensure that there is only one sink
        #  node for any global data in the combined state. The transients are handled
        #  naturally because we maintain the SSA invariant, furthermore, the inputs
        #  are already merged.
        # TODO(phimuell): Improve this merging.
        global_sink_nodes: dict[str, set[dace_nodes.AccessNode]] = {}
        for sink_node in first_state.sink_nodes():
            if not isinstance(sink_node, dace_nodes.AccessNode):
                continue
            if sink_node.desc(sdfg).transient:
                continue
            if sink_node.data not in global_sink_nodes:
                global_sink_nodes[sink_node.data] = set()
            global_sink_nodes[sink_node.data].add(sink_node)

        for sink_nodes in global_sink_nodes.values():
            if len(sink_nodes) <= 1:
                continue
            # We now select one node and redirect all writes to it.
            final_sink_node = sink_nodes.pop()
            for sink_node in sink_nodes:
                for iedge in first_state.in_edges(sink_node):
                    first_state.add_edge(
                        iedge.src, iedge.src_conn, final_sink_node, iedge.dst_conn, iedge.data
                    )
                first_state.remove_node(sink_node)

    def apply(self, _: Any, sdfg: dace.SDFG) -> None:
        first_state: dace.SDFGState = self.first_state
        second_state: dace.SDFGState = self.second_state
        graph: dace.sdfg.state.AbstractControlFlowRegion = first_state.parent_graph

        # Move the nodes from the second into the first state.
        self._move_nodes(sdfg)

        # Now redirect all state edges and remove the second state.
        # NOTE: It is important that we do not remove the first state. If we would
        #  remove it, then we would have to modify the start block property of the
        #  graph.
        dace_sdutils.change_edge_src(graph, second_state, first_state)
        graph.remove_node(second_state)
